/* This work is partially financed by the European Commission under the
* Framework 6 Information Society Technologies Project
* "Wirelessly Accessible Sensor Populations (WASP)".
*/

/*
GCC 4.x port by Ivan Shcherbakov <mspgcc@sysprogs.org>
*/

#include "config.h"
#include "system.h"
#include "coretypes.h"
#include "tm.h"
#include "rtl.h"
#include "regs.h"
#include "hard-reg-set.h"
#include "real.h"
#include "insn-config.h"
#include "conditions.h"
#include "insn-attr.h"
#include "flags.h"
#include "reload.h"
#include "tree.h"
#include "output.h"
#include "expr.h"
#include "toplev.h"
#include "obstack.h"
#include "function.h"
#include "recog.h"
#include "tm_p.h"
#include "target.h"
#include "target-def.h"
#include "insn-codes.h"
#include "ggc.h"
#include "langhooks.h"
#include "rtl-error.h"

/* The following functions are defined in this file and used by msp430.c */
void msp430_init_builtins (void);
rtx msp430_expand_builtin (tree, rtx, rtx, enum machine_mode, int);

enum msp430_builtins
{
  MSP430_BUILTIN_NOP,
  MSP430_BUILTIN_DINT,
  MSP430_BUILTIN_EINT,
  MSP430_BUILTIN_READ_STATUS_REGISTER,
  MSP430_BUILTIN_WRITE_STATUS_REGISTER,
  MSP430_BUILTIN_BIC_STATUS_REGISTER,
  MSP430_BUILTIN_BIS_STATUS_REGISTER,
  MSP430_BUILTIN_BIC_SR_IRQ,
  MSP430_BUILTIN_BIS_SR_IRQ,
  MSP430_BUILTIN_READ_STACK_POINTER,
  MSP430_BUILTIN_WRITE_STACK_POINTER,
  MSP430_BUILTIN_DELAY_CYCLES,
  MSP430_BUILTIN_GET_INTERRUPT_STATE,
  MSP430_BUILTIN_SET_INTERRUPT_STATE,
  MSP430_BUILTIN_SWAP_BYTES,
  MSP430_BUILTIN_GET_WATCHDOG_CLEAR_VALUE,
  MSP430_BUILTIN_SET_WATCHDOG_CLEAR_VALUE,
  MSP430_BUILTIN_WATCHDOG_CLEAR,
  MSP430_BUILTIN_last_enum
};

void
msp430_init_builtins (void)
{
  add_builtin_function ("__nop",
			build_function_type_list (void_type_node, NULL_TREE),
			MSP430_BUILTIN_NOP, BUILT_IN_MD, NULL, NULL_TREE);
  add_builtin_function ("__dint",
			build_function_type_list (void_type_node, NULL_TREE),
			MSP430_BUILTIN_DINT, BUILT_IN_MD, NULL, NULL_TREE);
  add_builtin_function ("__eint",
			build_function_type_list (void_type_node, NULL_TREE),
			MSP430_BUILTIN_EINT, BUILT_IN_MD, NULL, NULL_TREE);
  add_builtin_function ("__read_status_register",
			build_function_type_list (unsigned_type_node,
						  NULL_TREE),
			MSP430_BUILTIN_READ_STATUS_REGISTER, BUILT_IN_MD,
			NULL, NULL_TREE);
  add_builtin_function ("__get_interrupt_state",
			build_function_type_list (unsigned_type_node,
						  NULL_TREE),
			MSP430_BUILTIN_GET_INTERRUPT_STATE, BUILT_IN_MD, NULL,
			NULL_TREE);
  add_builtin_function ("__write_status_register",
			build_function_type_list (void_type_node,
						  unsigned_type_node,
						  NULL_TREE),
			MSP430_BUILTIN_WRITE_STATUS_REGISTER, BUILT_IN_MD,
			NULL, NULL_TREE);
  add_builtin_function ("__bic_status_register",
			build_function_type_list (void_type_node,
						  unsigned_type_node,
						  NULL_TREE),
			MSP430_BUILTIN_BIC_STATUS_REGISTER, BUILT_IN_MD, NULL,
			NULL_TREE);
  add_builtin_function ("__bis_status_register",
			build_function_type_list (void_type_node,
						  unsigned_type_node,
						  NULL_TREE),
			MSP430_BUILTIN_BIS_STATUS_REGISTER, BUILT_IN_MD, NULL,
			NULL_TREE);
  add_builtin_function ("__set_interrupt_state",
			build_function_type_list (void_type_node,
						  unsigned_type_node,
						  NULL_TREE),
			MSP430_BUILTIN_SET_INTERRUPT_STATE, BUILT_IN_MD, NULL,
			NULL_TREE);
  add_builtin_function ("__bic_status_register_on_exit",
			build_function_type_list (void_type_node,
						  unsigned_type_node,
						  NULL_TREE),
			MSP430_BUILTIN_BIC_SR_IRQ, BUILT_IN_MD, NULL,
			NULL_TREE);
  add_builtin_function ("__bis_status_register_on_exit",
			build_function_type_list (void_type_node,
						  unsigned_type_node,
						  NULL_TREE),
			MSP430_BUILTIN_BIS_SR_IRQ, BUILT_IN_MD, NULL,
			NULL_TREE);
  add_builtin_function ("__read_stack_pointer",
			build_function_type_list (ptr_type_node, NULL_TREE),
			MSP430_BUILTIN_READ_STACK_POINTER, BUILT_IN_MD, NULL,
			NULL_TREE);
  add_builtin_function ("__write_stack_pointer",
			build_function_type_list (void_type_node,
						  ptr_type_node, NULL_TREE),
			MSP430_BUILTIN_WRITE_STACK_POINTER, BUILT_IN_MD, NULL,
			NULL_TREE);
  add_builtin_function ("__delay_cycles",
			build_function_type_list (void_type_node,
						  long_unsigned_type_node,
						  NULL_TREE),
			MSP430_BUILTIN_DELAY_CYCLES, BUILT_IN_MD, NULL,
			NULL_TREE);
  add_builtin_function ("__swap_bytes",
			build_function_type_list (unsigned_type_node,
						  unsigned_type_node,
						  NULL_TREE),
			MSP430_BUILTIN_SWAP_BYTES, BUILT_IN_MD, NULL,
			NULL_TREE);
  add_builtin_function ("__get_watchdog_clear_value",
			build_function_type_list (unsigned_type_node, NULL_TREE),
			MSP430_BUILTIN_GET_WATCHDOG_CLEAR_VALUE, BUILT_IN_MD, NULL,
			NULL_TREE);
  add_builtin_function ("__set_watchdog_clear_value",
			build_function_type_list (void_type_node,
						  unsigned_type_node, NULL_TREE),
			MSP430_BUILTIN_SET_WATCHDOG_CLEAR_VALUE, BUILT_IN_MD, NULL,
			NULL_TREE);
  add_builtin_function ("__watchdog_clear",
			build_function_type_list (void_type_node, NULL_TREE),
			MSP430_BUILTIN_WATCHDOG_CLEAR, BUILT_IN_MD, NULL, NULL_TREE);
}

rtx
msp430_expand_builtin (tree exp, rtx target ATTRIBUTE_UNUSED,
		       rtx subtarget ATTRIBUTE_UNUSED,
		       enum machine_mode mode ATTRIBUTE_UNUSED,
		       int ignore ATTRIBUTE_UNUSED)
{
  rtx arg = 0;
  rtx retval = 0;
  rtx deref_stack_sr = 0;
  rtx insn = 0;
  bool need_insn = true;
  tree fndecl;
  enum msp430_builtins builtin_code;
  struct machine_function *mfp = cfun->machine;

  fndecl = TREE_OPERAND (CALL_EXPR_FN (exp), 0);
  builtin_code = (enum msp430_builtins) DECL_FUNCTION_CODE (fndecl);

  if (builtin_code == MSP430_BUILTIN_BIC_SR_IRQ
      || builtin_code == MSP430_BUILTIN_BIS_SR_IRQ)
    {
      if (mfp->interrupt == NULL_TREE)
	error
	  ("__bi%c_status_register_on_exit invalid outside of function with interrupt attribute",
	   (builtin_code == MSP430_BUILTIN_BIC_SR_IRQ) ? 'c' : 's');
      /* If we used hard_frame_pointer_rtx here, which would be right,
         we'd end up setting frame_pointer_required even if it isn't.
         So use arg_pointer_rtx offset by the two words pushed by the
         interrupt. */
      deref_stack_sr =
	gen_rtx_MEM (Pmode,
		     plus_constant (arg_pointer_rtx, -2 * UNITS_PER_WORD));
    }

  switch (builtin_code)
    {
    default:
      break;
    case MSP430_BUILTIN_NOP:
      insn = gen_nop ();
      break;
    case MSP430_BUILTIN_DINT:
      emit_insn (gen_dint ());
      insn = gen_nop ();
      break;
    case MSP430_BUILTIN_EINT:
      insn = gen_eint ();
      break;
    case MSP430_BUILTIN_READ_STATUS_REGISTER:
      retval = gen_reg_rtx (HImode);
      insn = gen_read_status_register (retval);
      break;
    case MSP430_BUILTIN_WRITE_STATUS_REGISTER:
      arg =
	expand_expr (CALL_EXPR_ARG (exp, 0), NULL_RTX, VOIDmode,
		     EXPAND_NORMAL);
      insn = gen_write_status_register (arg);
      break;
    case MSP430_BUILTIN_GET_INTERRUPT_STATE:
      retval = gen_reg_rtx (HImode);
      insn = gen_get_interrupt_state (retval);
      break;
    case MSP430_BUILTIN_BIC_STATUS_REGISTER:
      arg =
	expand_expr (CALL_EXPR_ARG (exp, 0), NULL_RTX, VOIDmode,
		     EXPAND_NORMAL);
      insn = gen_bic_status_register (arg);
      break;
    case MSP430_BUILTIN_BIS_STATUS_REGISTER:
      arg =
	expand_expr (CALL_EXPR_ARG (exp, 0), NULL_RTX, VOIDmode,
		     EXPAND_NORMAL);
      insn = gen_bis_status_register (arg);
      break;
    case MSP430_BUILTIN_SET_INTERRUPT_STATE:
      arg =
	expand_expr (CALL_EXPR_ARG (exp, 0), NULL_RTX, VOIDmode,
		     EXPAND_NORMAL);
      insn = gen_set_interrupt_state (arg);
      break;
    case MSP430_BUILTIN_BIC_SR_IRQ:
      arg =
	expand_expr (CALL_EXPR_ARG (exp, 0), NULL_RTX, VOIDmode,
		     EXPAND_NORMAL);
      insn = gen_bic_status_register_on_exit (deref_stack_sr, arg);
      break;
    case MSP430_BUILTIN_BIS_SR_IRQ:
      arg =
	expand_expr (CALL_EXPR_ARG (exp, 0), NULL_RTX, VOIDmode,
		     EXPAND_NORMAL);
      insn = gen_bis_status_register_on_exit (deref_stack_sr, arg);
      break;
    case MSP430_BUILTIN_READ_STACK_POINTER:
      retval = gen_reg_rtx (HImode);
      insn = gen_read_stack_pointer (retval);
      break;
    case MSP430_BUILTIN_WRITE_STACK_POINTER:
      arg =
	expand_expr (CALL_EXPR_ARG (exp, 0), NULL_RTX, VOIDmode,
		     EXPAND_NORMAL);
      insn = gen_write_stack_pointer (arg);
      break;
    case MSP430_BUILTIN_GET_WATCHDOG_CLEAR_VALUE:
      retval = gen_reg_rtx (HImode);
      if (msp430_disable_watchdog)
	insn = gen_movhi (retval, constm1_rtx);
      else
	insn = gen_get_watchdog_clear_value (retval);
      break;
    case MSP430_BUILTIN_SET_WATCHDOG_CLEAR_VALUE:
      if (msp430_disable_watchdog)
	need_insn = false;
      else
	{
	  arg =
	    expand_expr (CALL_EXPR_ARG (exp, 0), NULL_RTX, VOIDmode,
			 EXPAND_NORMAL);
	  insn = gen_set_watchdog_clear_value (arg);
	}
      break;
    case MSP430_BUILTIN_WATCHDOG_CLEAR:
      if (msp430_disable_watchdog)
	need_insn = false;
      else
	insn = gen_watchdog_clear ();
      break;
    case MSP430_BUILTIN_DELAY_CYCLES:
      {
	tree cycles_arg = CALL_EXPR_ARG (exp, 0);
	const enum machine_mode itercnt_mode = HImode;
	const HOST_WIDE_INT itercnt_max =
	  (HOST_WIDE_INT) 1 << GET_MODE_BITSIZE (itercnt_mode);
	const int decr_cycles = 1;
	const int jump_cycles = 2;
	const int iter_overhead_cycles = decr_cycles + jump_cycles;
	const int iter0_cycles = iter_overhead_cycles;
	const int cg_init_cycles = 1;
	const int noncg_init_cycles = 2;
	const HOST_WIDE_INT iter1_cycles =
	  itercnt_max * iter0_cycles + iter_overhead_cycles;
	HOST_WIDE_INT cycles;
	rtx iterreg0 = NULL_RTX;

	need_insn = false;
	if (!cst_and_fits_in_hwi (cycles_arg)
	    || (0 > ((cycles = int_cst_value (cycles_arg)))))
	  {
	    error
	      ("__delay_cycles argument must be non-negative integer constant");
	    break;
	  }
	while (0 < cycles)
	  {
	    HOST_WIDE_INT itercnt1 = 0;
	    HOST_WIDE_INT itercnt0;

	    if (cycles >= iter_overhead_cycles)
	      itercnt1 = (cycles - iter_overhead_cycles) / iter1_cycles;
	    if (itercnt1 > 0)
	      itercnt0 =
		(cycles - itercnt1 * iter1_cycles -
		 iter_overhead_cycles) / iter0_cycles;
	    else
	      itercnt0 = cycles / iter0_cycles;
	    if (itercnt0 > itercnt_max)
	      {
		itercnt0 -= itercnt_max;
		++itercnt1;
	      }
	    else if ((itercnt1 > 0) && (itercnt0 == 0))
	      {
		itercnt1 -= 1;
		itercnt0 += itercnt_max;
	      }
	    gcc_assert (0 <= itercnt0 && itercnt0 <= itercnt_max);
	    if (itercnt0 > 0)
	      {
		while (1)
		  {
		    unsigned HOST_WIDE_INT loop_cycles;
		    int cycles_in_init;

		    cycles_in_init =
		      MSP430_CG_INT_P (trunc_int_for_mode
				       (itercnt0,
					itercnt_mode)) ? cg_init_cycles :
		      noncg_init_cycles;
		    loop_cycles = itercnt0 * iter0_cycles;
		    if (itercnt1 > 0)
		      {
			loop_cycles +=
			  itercnt1 * iter1_cycles + iter_overhead_cycles;
			cycles_in_init +=
			  MSP430_CG_INT_P (trunc_int_for_mode
					   (1 + itercnt1,
					    itercnt_mode)) ? cg_init_cycles :
			  noncg_init_cycles;
		      }
		    if (loop_cycles + cycles_in_init <= (unsigned HOST_WIDE_INT) cycles)
		      break;

		    --itercnt0;
		    if (itercnt0 == 0 && itercnt1 > 0)
		      {
			itercnt0 += itercnt_max;
			--itercnt1;
		      }
		  }
		if (itercnt0 > 0)
		  {
		    rtx loop_label = gen_label_rtx ();
		    rtx loop_test;
		    rtx iterreg1 = NULL_RTX;

		    if (itercnt1 > 0)
		      iterreg1 = gen_reg_rtx (itercnt_mode);
		    if (iterreg0 == NULL_RTX)
		      iterreg0 = gen_reg_rtx (itercnt_mode);

		    if (iterreg1 != NULL_RTX)
		      {
			emit_insn (gen_delay_cycles_init
				   (iterreg1,
				    gen_int_mode (1 + itercnt1,
						  itercnt_mode)));
			cycles -=
			  MSP430_CG_INT_P (trunc_int_for_mode
					   (1 + itercnt1,
					    itercnt_mode)) ? cg_init_cycles :
			  noncg_init_cycles;
		      }

		    emit_insn (gen_delay_cycles_init
			       (iterreg0,
				gen_int_mode (itercnt0, itercnt_mode)));
		    cycles -=
		      MSP430_CG_INT_P (trunc_int_for_mode
				       (itercnt0,
					itercnt_mode)) ? cg_init_cycles :
		      noncg_init_cycles;

		    emit_label (loop_label);
		    emit_insn (gen_delay_cycles_decr (iterreg0));
		    loop_test = gen_rtx_NE (itercnt_mode, iterreg0, const0_rtx);
		    gcc_assert (HImode == itercnt_mode);
		    emit_jump_insn (gen_cbranchhi4 (loop_test, iterreg0, const0_rtx, loop_label));
		    cycles -= itercnt0 * iter0_cycles;
		    if (iterreg1 != NULL_RTX)
		      {
			emit_insn (gen_delay_cycles_decr (iterreg1));
			loop_test = gen_rtx_NE (itercnt_mode, iterreg1, const0_rtx);
			emit_jump_insn (gen_cbranchhi4 (loop_test, iterreg1, const0_rtx, loop_label));
			cycles -=
			  itercnt1 * iter1_cycles + iter_overhead_cycles;
		      }
		    /* In the areas between single (max 196608) and
		     * double (min 196619) loops, the remainder may
		     * warrant another single loop instead of a
		     * sequence of nine nops. */
		    gcc_assert (0 <= cycles);
		    continue;
		  }
	      }
	    while (cycles--)
	      emit_insn (gen_nop ());
	    break;
	  }
	break;
      }
    case MSP430_BUILTIN_SWAP_BYTES:
      retval = gen_reg_rtx (HImode);
      arg =
	expand_expr (CALL_EXPR_ARG (exp, 0), NULL_RTX, VOIDmode,
		     EXPAND_NORMAL);
      emit_move_insn (retval, arg);
      insn = gen_bswaphi1 (retval);
      break;
    }

  if (insn)
    emit_insn (insn);
  else if (need_insn)
    error ("Unhandled built-in function `%s'",
	   IDENTIFIER_POINTER (DECL_NAME (fndecl)));

  return retval;
}
