/* TODO
 * cost speed attribute
 */
/* This work is partially financed by the European Commission under the
* Framework 6 Information Society Technologies Project
* "Wirelessly Accessible Sensor Populations (WASP)".
*/

/*
	GCC 4.x port by Ivan Shcherbakov <mspgcc@sysprogs.org>
*/

/* Subroutines for insn-output.c for Texas Instruments MSP430 MCU
Copyright (C) 2001, 2002 Free Software Foundation, Inc.
Contributed by Dmitry Diky <diwil@mail.ru>

This file is part of GNU CC. 
GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include "config.h"
#include "system.h"
#include "coretypes.h"
#include "tm.h"
#include "rtl.h"
#include "regs.h"
#include "hard-reg-set.h"
#include "real.h"
#include "insn-config.h"
#include "conditions.h"
#include "insn-attr.h"
#include "flags.h"
#include "reload.h"
#include "tree.h"
#include "output.h"
#include "expr.h"
#include "toplev.h"
#include "obstack.h"
#include "function.h"
#include "optabs.h"
#include "recog.h"
#include "tm_p.h"
#include "target.h"
#include "target-def.h"
#include "insn-codes.h"
#include "ggc.h"
#include "langhooks.h"
#include "df.h"
#include "intl.h"
#include "diagnostic.h"

/* Number of consecutive registers available for returning values from
 * a function.  These start with MSP430_RETURN_REGISTER_BASE and grow
 * "down". */
#define RETURN_REGISTERS_AVAILABLE 4

msp430_cpu_e msp430_cpu = MSP430_CPU_MSP430;
msp430_mpy_e msp430_mpy = MSP430_MPY_NONE;
int msp430_ivcnt = 16;

int msp430_case_values_threshold = 30000;
int msp430_has_hwmul = 0;

/* Defined in msp430-builtins.c */
void msp430_init_builtins (void);
rtx msp430_expand_builtin (tree, rtx, rtx, enum machine_mode, int);

static struct machine_function *
msp430_init_machine_status (void)
{
  return ggc_alloc_cleared_machine_function ();
}

static GTY(()) rtx mpy_b_rtx;
static GTY(()) rtx mpys_b_rtx;
static GTY(()) rtx op2_b_rtx;
static GTY(()) rtx reslo_b_rtx;
static GTY(()) rtx mpy_rtx;
static GTY(()) rtx mpys_rtx;
static GTY(()) rtx op2_rtx;
static GTY(()) rtx reslo_rtx;
static GTY(()) rtx reshi_rtx;
static GTY(()) rtx mpy32l_rtx;
static GTY(()) rtx mpy32h_rtx;
static GTY(()) rtx mpys32l_rtx;
static GTY(()) rtx mpys32h_rtx;
static GTY(()) rtx op2l_rtx;
static GTY(()) rtx op2h_rtx;
static GTY(()) rtx res0_rtx;
static GTY(()) rtx res1_rtx;
static GTY(()) rtx res2_rtx;
static GTY(()) rtx res3_rtx;

static GTY(()) rtx libsym_mulqi3;
static GTY(()) rtx libsym_mulqihi3;
static GTY(()) rtx libsym_umulqihi3;
static GTY(()) rtx libsym_mulhi3;
static GTY(()) rtx libsym_mulhisi3;
static GTY(()) rtx libsym_umulhisi3;
static GTY(()) rtx libsym_mulsi3;
static GTY(()) rtx libsym_mulsidi3;
static GTY(()) rtx libsym_umulsidi3;
static GTY(()) rtx libsym_muldi3;

static GTY(()) rtx libsym_ashlqi3;
static GTY(()) rtx libsym_ashrqi3;
static GTY(()) rtx libsym_lshrqi3;
static GTY(()) rtx libsym_ashlhi3;
static GTY(()) rtx libsym_ashrhi3;
static GTY(()) rtx libsym_lshrhi3;
static GTY(()) rtx libsym_ashlsi3;
static GTY(()) rtx libsym_ashrsi3;
static GTY(()) rtx libsym_lshrsi3;
static GTY(()) rtx libsym_ashldi3;
static GTY(()) rtx libsym_ashrdi3;
static GTY(()) rtx libsym_lshrdi3;

static rtx
gen_rtx_HWREG (enum machine_mode mode, const char *name)
{
  rtx ret = gen_rtx_MEM (mode, gen_rtx_SYMBOL_REF (mode, name));
  MEM_VOLATILE_P (ret) = 1;
  return ret;
}

static void
msp430_init_once (void)
{
  init_machine_status = msp430_init_machine_status;

  if (msp430_mpy & MSP430_MPY_TYPE_16)
    {
      mpy_b_rtx = gen_rtx_HWREG (QImode, "__MPY");
      mpys_b_rtx = gen_rtx_HWREG (QImode, "__MPYS");
      op2_b_rtx = gen_rtx_HWREG (QImode, "__OP2");
      reslo_b_rtx = gen_rtx_HWREG (QImode, "__RESLO");
      mpy_rtx = gen_rtx_HWREG (HImode, "__MPY");
      mpys_rtx = gen_rtx_HWREG (HImode, "__MPYS");
      op2_rtx = gen_rtx_HWREG (HImode, "__OP2");
      reslo_rtx = gen_rtx_HWREG (HImode, "__RESLO");
      reshi_rtx = gen_rtx_HWREG (HImode, "__RESHI");
    }
  if (msp430_mpy & MSP430_MPY_TYPE_32)
    {
      mpy32l_rtx = gen_rtx_HWREG (HImode, "__MPY32L");
      mpy32h_rtx = gen_rtx_HWREG (HImode, "__MPY32H");
      mpys32l_rtx = gen_rtx_HWREG (HImode, "__MPYS32L");
      mpys32h_rtx = gen_rtx_HWREG (HImode, "__MPYS32H");
      op2l_rtx = gen_rtx_HWREG (HImode, "__OP2L");
      op2h_rtx = gen_rtx_HWREG (HImode, "__OP2H");
      res0_rtx = gen_rtx_HWREG (HImode, "__RES0");
      res1_rtx = gen_rtx_HWREG (HImode, "__RES1");
      res2_rtx = gen_rtx_HWREG (HImode, "__RES2");
      res3_rtx = gen_rtx_HWREG (HImode, "__RES3");
    }
  return;
}

/****** ATTRIBUTES *************************************/

/* Handle an attribute requiring a FUNCTION_DECL; arguments as in
struct attribute_spec.handler.  */
static tree
handle_fndecl_attribute (tree * pnode, tree name,
			 tree args ATTRIBUTE_UNUSED,
			 int flags ATTRIBUTE_UNUSED, bool * no_add_attrs)
{
  if (TREE_CODE (*pnode) != FUNCTION_DECL)
    {
      warning (OPT_Wattributes, "%qE attribute ignored", name);
      *no_add_attrs = true;
    }
  return NULL_TREE;
}


const struct attribute_spec msp430_attribute_table[] = {
  /* { name, min_len, max_len, decl_req, type_req, fn_type_req, handler } */
  {"signal", 0, 0, true, false, false, handle_fndecl_attribute},
  {"interrupt", 0, 1, true, false, false, handle_fndecl_attribute},
  {"naked", 0, 0, true, false, false, handle_fndecl_attribute},
  {"task", 0, 0, true, false, false, handle_fndecl_attribute},
  {"wakeup", 0, 0, true, false, false, handle_fndecl_attribute},
  {"critical", 0, 0, true, false, false, handle_fndecl_attribute},
  {"reentrant", 0, 0, true, false, false, handle_fndecl_attribute},
  {"saveprologue", 0, 0, true, false, false, handle_fndecl_attribute},
  {"noint_hwmul", 0, 0, true, false, false, handle_fndecl_attribute},
  {"hosted", 0, 0, true, false, false, handle_fndecl_attribute},
  {NULL, 0, 0, false, false, false, NULL}
};

int
msp430_current_function_noint_hwmul_function_p (void)
{
  return TARGET_NOINT_HWMUL || (cfun && cfun->machine
				&& cfun->machine->initialized
				&& cfun->machine->noint_hwmul != NULL_TREE);
}

unsigned int
msp430_section_type_flags (tree decl, const char *name, int reloc)
{
  unsigned int flags = default_section_type_flags (decl, name, reloc);
  const char *suffix = strrchr (name, '.');

  if (!strcmp (name, ".noinit")
      || !strcmp (name, ".infomemnobits")
      || (suffix && 0 == strcmp (suffix, ".bss")))
    {
      if (!strcmp (name, ".infomemnobits"))
	warning_at (DECL_SOURCE_LOCATION (decl), 0,
		    _(".infomemnobits is deprecated; use .infomem.bss"));
      if (!decl
	  || (TREE_CODE (decl) == VAR_DECL
	      && DECL_INITIAL (decl) == NULL_TREE))
	flags |= SECTION_BSS;	/* @nobits */
      else
	warning_at (DECL_SOURCE_LOCATION (decl), 0,
		    "only uninitialized variables can be placed in a .bss section");
    }
  return flags;
}

struct tag_value_pair_t
{
  const char *tag;
  unsigned long value;
};

static const struct tag_value_pair_t cpu_tag_value_map[] = {
  {"430", MSP430_CPU_MSP430},
  {"430x", MSP430_CPU_MSP430X},
  {"430xv2", MSP430_CPU_MSP430XV2},
  {0, 0}
};

static const struct tag_value_pair_t mpy_tag_value_map[] = {
  {"none", MSP430_MPY_NONE},
  {"16", MSP430_MPY_16},
  {"16se", MSP430_MPY_16SE},
  {"32", MSP430_MPY_32},
  {"32dw", MSP430_MPY_32DW},
  {0, 0}
};

static const struct tag_value_pair_t *
find_pair_by_tag (const char *tag, const struct tag_value_pair_t *map)
{
  while (map->tag)
    {
      if (0 == strcmp (tag, map->tag))
	return map;
      ++map;
    }
  return 0;
}

static const struct tag_value_pair_t *
find_pair_by_value (unsigned long value, const struct tag_value_pair_t *map)
{
  while (map->tag)
    {
      if (map->value == value)
	return map;
      ++map;
    }
  return 0;
}

static void
msp430_option_override (void)
{
  const struct tag_value_pair_t *mp;

  if (flag_pic)
    error ("PIC not supported on msp430");

  if (msp430_mcu_name)
    {
      char * up = strchr(msp430_mcu_name, '_');
      if (NULL != up)
	*up = 0;
    }

  if (msp430_opt_cpu)
    {
      mp = find_pair_by_tag (msp430_opt_cpu, cpu_tag_value_map);
      if (!mp)
	{
	  error (_("mcpu %s not recognized"), msp430_opt_cpu);
	  return;
	}
      msp430_cpu = (msp430_cpu_e) mp->value;
    }

  if (msp430_opt_mpy)
    {
      mp = find_pair_by_tag (msp430_opt_mpy, mpy_tag_value_map);
      if (!mp)
	{
	  error (_("mmpy %s not recognized"), msp430_opt_mpy);
	  return;
	}
      msp430_mpy = (msp430_mpy_e) mp->value;
    }

  if (msp430_opt_ivcnt)
    {
      char *ep;
      msp430_ivcnt = strtol (msp430_opt_ivcnt, &ep, 0);
      if (*msp430_opt_ivcnt && *ep)
	{
	  error (_("ivcnt %s must be integer"), msp430_opt_ivcnt);
	  gcc_unreachable ();
	  return;
	}
      if ((16 != msp430_ivcnt) && (32 != msp430_ivcnt)
	  && (64 != msp430_ivcnt))
	{
	  error (_("ivcnt %s must be 16, 32, or 64"), msp430_opt_ivcnt);
	  gcc_unreachable ();
	  return;
	}
    }

  if (TARGET_NVWA)
    warning (OPT_Wdeprecated, _("-mno-volatile-workaround ignored"));

  if (TARGET_NO_HWMUL)
    {
      warning (OPT_Wdeprecated, "-mdisable-hwmul deprecated, use -mmpy=none");
      msp430_mpy = MSP430_MPY_NONE;
    }
  msp430_has_hwmul = ! !(msp430_mpy & MSP430_MPY_TYPE_ANY);

  msp430_case_values_threshold = 8;	/* ? or there is a better value ? */

  flag_defer_pop = 0;

  msp430_init_once ();
}

/* Return in memory unless size is determined to be no greater than
 * eight bytes. */
static bool
msp430_return_in_memory (const_tree type, const_tree fntype ATTRIBUTE_UNUSED)
{
  if (TYPE_MODE (type) == BLKmode)
    {
      HOST_WIDE_INT size = int_size_in_bytes (type);
      return (size == -1
	      || size > (RETURN_REGISTERS_AVAILABLE * UNITS_PER_WORD));
    }
  return false;
}

static int reg_class_tab[FIRST_VIRTUAL_REGISTER] = {
  PC_REG, SP_REG, SR_REG, CG2_REG,	/* fixed registers */
  FP_REG,			/* maybe fixed reg */
  GENERAL_REGS, GENERAL_REGS, GENERAL_REGS,	/* unassigned registers r5 - r15 */
  GENERAL_REGS, GENERAL_REGS, GENERAL_REGS, GENERAL_REGS,
  GENERAL_REGS, GENERAL_REGS, GENERAL_REGS, GENERAL_REGS,
  SOFT_ARGP_REG, SOFT_FP_REG,	/* fixed pseudo registers */
};

int
msp430_regno_ok_for_base_p (int reg, int strict)
{
  if (strict && reg >= FIRST_PSEUDO_REGISTER && reg_renumber)
    reg = reg_renumber[reg];
  if (reg < 0)
    return 0;
  if (!strict && reg > MSP430_MAX_GENERAL_REGNUM)
    return 1;
  return reg == STACK_POINTER_REGNUM || MSP430_GENERAL_REGISTER_NUM_P (reg);
}

enum reg_class
msp430_regno_reg_class (int r)
{
  if (r < (int) (sizeof (reg_class_tab) / sizeof (reg_class_tab[0])))
    return (enum reg_class) reg_class_tab[r];

  return NO_REGS;
}

void
msp430_asm_declare_function_name (FILE * file, const char *name, tree decl)
{
  struct machine_function *mfp = cfun->machine;
  bool is_isr = NULL_TREE != mfp->interrupt;
  int vector = -1;

  if (is_isr && (0 <= mfp->vector_offset))
    {
      vector = mfp->vector_offset / 2;
      if (0 < msp430_ivcnt && (vector >= msp430_ivcnt))
	{
	  error_at (DECL_SOURCE_LOCATION (decl),
		    _
		    ("interrupt vector %d is beyond end of MCU vector table"),
		    vector);
	  vector = -1;
	  is_isr = false;
	}
    }

  fprintf (file, "%s", TYPE_ASM_OP);
  assemble_name (file, name);
  putc (',', file);
  fprintf (file, TYPE_OPERAND_FMT, "function");
  putc ('\n', file);
  fprintf (file, "/***********************\n");
  if (!is_isr)
    fprintf (file, " * Function `");
  else if (0 > vector)
    fprintf (file, " * Interrupt Sub-Routine `");
  else
    fprintf (file, " * Interrupt Vector %d Service Routine `", vector);
  assemble_name (file, name);
  fprintf (file, "' %s\n ***********************/\n",
	   (cfun->machine->critical != NULL_TREE)
	   ? "(OS critical)"
	   : (cfun->machine->reentrant != NULL_TREE) ? "(reentrant)" : "");
  ASM_OUTPUT_LABEL (file, name);
  if (0 <= vector)
    {
      char isrname[32];
      snprintf (isrname, sizeof (isrname), "__isr_%u", vector);
      fprintf (file, ".global\t%s\n", isrname);
      fprintf (file, "%s:\n", isrname);
    }

}

static bool
msp430_legitimate_address_p (enum machine_mode mode ATTRIBUTE_UNUSED,
			     rtx x, bool strict)
{
  switch (GET_CODE (x))
    {
    case REG:
      return msp430_regno_ok_for_base_p (REGNO (x), strict);
    case PLUS:
      return REG_P (XEXP (x, 0))
	&& msp430_legitimate_address_p (mode, XEXP (x, 0), strict)
	&& CONSTANT_P (XEXP (x, 1));
    default:
      return CONSTANT_P (x);
    }
  gcc_unreachable ();
}

static const char *trim_array[] = { "llo", "lhi", "hlo", "hhi" };

static void
print_sub_operand (FILE * file, rtx x, int code)
{

  if (GET_CODE (x) == SYMBOL_REF || GET_CODE (x) == LABEL_REF)
    output_addr_const (file, x);
  else if (GET_CODE (x) == CONST)
    print_sub_operand (file, XEXP (x, 0), code);
  else if (GET_CODE (x) == PLUS)
    {
      print_sub_operand (file, XEXP (x, 0), code);
      fprintf (file, "+");
      print_sub_operand (file, XEXP (x, 1), code);
    }
  else if (GET_CODE (x) == CONST_INT)
    fprintf (file, "%ld", INTVAL (x));
  else
    gcc_unreachable ();
}

void
msp430_print_operand (FILE * file, rtx x, int code)
{
  int shift = 0;

  if (code >= 'A' && code <= 'D')
    shift = code - 'A';
  else if (code != 0 && code != 'S')
    gcc_unreachable ();

  if (REG_P (x))
    fprintf (file, reg_names[REGNO (x) + shift]);
  else if (GET_CODE (x) == CONST_INT)
    {
      HOST_WIDE_INT intval = INTVAL (x);

      /* For improved ASM readability, omit #llo(const) for small constants */
      if (!shift && !(intval & ~0xFFFF))
	fprintf (file, "#%d", (int) intval);
      else
	fprintf (file, "#%s(%ld)", trim_array[shift], intval);
    }
  else if (GET_CODE (x) == MEM)
    {
      rtx addr = XEXP (x, 0);

      if (GET_CODE (addr) == POST_INC)
	fprintf (file, "@%s+", reg_names[REGNO (XEXP (addr, 0))]);
      else if (GET_CODE (addr) == REG)
	{			/* for X(Rn) */
	  int regno = REGNO (addr);
	  const char* regname = reg_names[regno];
	  
	  if (code == 'S' && regno == STACK_POINTER_REGNUM)
	    fprintf (file, "%d(%s)", UNITS_PER_WORD, regname);
	  else if (shift)
	    fprintf (file, "%d(%s)", shift * UNITS_PER_WORD, regname);
	  else
	    fprintf (file, "@%s", regname);
	}
      else if (GET_CODE (addr) == SYMBOL_REF)
	{
	  fprintf (file, "&");
	  output_addr_const (file, addr);
	  if (shift)
	    fprintf (file, "+%d", shift * UNITS_PER_WORD);
	}
      else if (GET_CODE (addr) == CONST || GET_CODE (addr) == CONST_INT)
	{
	  fputc ('&', file);
	  output_addr_const (file, addr);
	  if (shift)
	    fprintf (file, "+%d", shift * UNITS_PER_WORD);
	}
      else if (GET_CODE (addr) == PLUS)
	{
	  rtx base = XEXP (addr, 0);
	  rtx offset = XEXP (addr, 1);
	  int regno = REGNO (base);
	  const char* regname = reg_names[regno];

	  gcc_assert (REG_P (base));
	  print_sub_operand (file, offset, code);
	  if (code == 'S' && regno == STACK_POINTER_REGNUM)
	    fprintf (file, "+%d(%s)", UNITS_PER_WORD, regname);
	  else
	    {
	      if (shift)
		fprintf (file, "+%d", shift * UNITS_PER_WORD);
	      fprintf (file, "(%s)", regname);
	    }
	}
      else
	gcc_unreachable ();
    }
  else if (GET_CODE (x) == SYMBOL_REF || GET_CODE (x) == LABEL_REF || GET_CODE (x) == CONST)
    {
      fprintf (file, "#");
      output_addr_const (file, x);
      if (shift)
	fprintf (file, "+%d", shift * UNITS_PER_WORD);
    }
  else if (GET_CODE (x) == CODE_LABEL)
    output_addr_const (file, x);
  else if (GET_CODE (x) == CONST_DOUBLE)
    {
      if (GET_MODE (x) == VOIDmode)	/* FIXME: may be long long?? */
	{
	  if (shift < 2)
	    fprintf (file, "#%s(%ld)", trim_array[shift],
		     CONST_DOUBLE_LOW (x));
	  else
	    fprintf (file, "#%s(%ld)", trim_array[shift - 2],
		     CONST_DOUBLE_HIGH (x));
	}
      else if (GET_MODE (x) == SFmode || GET_MODE (x) == SImode)
	{
	  long val;
	  REAL_VALUE_TYPE rv;
	  REAL_VALUE_FROM_CONST_DOUBLE (rv, x);
	  REAL_VALUE_TO_TARGET_SINGLE (rv, val);
	  asm_fprintf (file, "#%s(0x%lx)", trim_array[shift], val);
	}
      else
	gcc_unreachable ();
    }
  else
    gcc_unreachable ();
}

/* Convert an operand that requires multiple words into a sequence of
   word-sized operands comprising its elements, from least significant
   to most significant.  Returns the number of word elements
   extracted. */
int
msp430_extract_multiword_operand (enum machine_mode innermode, rtx op,
				  rtx * parts)
{
  int words;
  rtx last_insn = get_last_insn ();
  int i;

  if (VOIDmode == innermode)
    innermode = GET_MODE (op);
  words = GET_MODE_SIZE (innermode) / UNITS_PER_WORD;
  if (1 >= words)
    {
      parts[0] = op;
      words = 1;
    }
  else
    for (i = 0; i < words; ++i)
      {
	parts[i] =
	  simplify_gen_subreg (HImode, op, innermode, i * UNITS_PER_WORD);
	gcc_assert (parts[i]);
	if (MEM_P (parts[i]))
	  {
	    parts[i] = validize_mem (parts[i]);
	    gcc_assert (parts[i]);
	  }
      }
  gcc_assert (last_insn == get_last_insn ());
  return words;
}


/* Generate the appropriate template for a read-modify-write operation
 * in MODE, using the SRC_OPERAND as the operand index for a binary
 * pattern, with the least significant word generated using INIT_OP
 * and any higher words are generated using NEXT_OP (which defaults to
 * INIT_OP if NULL).  If SRC_OPERAND is 0, a unary pattern is used. */
static const char *
output_template (enum machine_mode mode,
		 int src_operand, bool reversep,
		 const char *init_op, const char *next_op,
		 const char *prefix)
{
  char buffer[1024];
  int offset = 0;

  if (!next_op)
    next_op = init_op;
  buffer[offset] = 0;
  if (prefix)
    offset += snprintf (buffer, sizeof (buffer) - offset,
			"%s\n\t", prefix);
  switch (mode)
    {
    case QImode:
      if (0 < src_operand)
	offset += snprintf (buffer + offset, sizeof (buffer) - offset,
			    "%s.b\t%%%d, %%0", init_op, src_operand);
      else
	offset += snprintf (buffer + offset, sizeof (buffer) - offset,
			    "%s.b\t%%0", init_op);
      break;
    case HImode:
      if (0 < src_operand)
	offset += snprintf (buffer + offset, sizeof (buffer) - offset,
			    "%s\t%%%d, %%0", init_op, src_operand);
      else
	offset += snprintf (buffer + offset, sizeof (buffer) - offset,
			    "%s\t%%0", init_op);
      break;
    case SImode:
    case SFmode:
      if (0 < src_operand)
	{
	  const char* format = reversep
	    ? "%s\t%%B%d, %%B0\n\t"
	      "%s\t%%A%d, %%A0"
	    : "%s\t%%A%d, %%A0\n\t"
	      "%s\t%%B%d, %%B0";

	  offset += snprintf (buffer + offset, sizeof (buffer) - offset,
			      format, init_op, src_operand, next_op, src_operand);
	}
      else
	{
	  const char* format = reversep
	    ? "%s\t%%B0\n\t" "%s\t%%A0"
	    : "%s\t%%A0\n\t" "%s\t%%B0";
	  offset += snprintf (buffer + offset, sizeof (buffer) - offset,
			      format, init_op, next_op);
	}
      break;
    case DImode:
      if (0 < src_operand)
	{
	  const char* format = reversep
	    ? "%s\t%%D%d, %%D0\n\t"
	      "%s\t%%C%d, %%C0\n\t"
	      "%s\t%%B%d, %%B0\n\t"
	      "%s\t%%A%d, %%A0"
	    : "%s\t%%A%d, %%A0\n\t"
	      "%s\t%%B%d, %%B0\n\t"
	      "%s\t%%C%d, %%C0\n\t"
 	      "%s\t%%D%d, %%D0";

	  offset += snprintf (buffer + offset, sizeof (buffer) - offset,
			      format,
			      init_op, src_operand,
			      next_op, src_operand,
			      next_op, src_operand, next_op, src_operand);
	}
      else
	{
	  const char* format = reversep
	    ? "%s\t%%D0\n\t"
	      "%s\t%%C0\n\t"
	      "%s\t%%B0\n\t"
	      "%s\t%%A0"
	    : "%s\t%%A0\n\t"
	      "%s\t%%B0\n\t"
	      "%s\t%%C0\n\t"
	      "%s\t%%D0";
	  offset += snprintf (buffer + offset, sizeof (buffer) - offset,
			      format, init_op, next_op, next_op, next_op);
	}
      break;
    default:
      gcc_unreachable ();
    }
  return ggc_strdup (buffer);
}

/* Generate the appropriate template for a read-modify-write operation
 * in MODE, using the SRC_OPERAND as the operand index for a binary
 * pattern, with the least significant word generated using INIT_OP
 * and any higher words are generated using NEXT_OP (which defaults to
 * INIT_OP if NULL).  If SRC_OPERAND is 0, a unary pattern is used. */
const char *
msp430_output_template (enum machine_mode mode,
			int src_operand,
			const char *init_op, const char *next_op,
			const char *prefix)
{
  return output_template (mode, src_operand, 0, init_op, next_op, prefix);
}

const char *
msp430_output_reverse_template (enum machine_mode mode,
				int src_operand,
				const char *init_op, const char *next_op,
				const char *prefix)
{
  return output_template (mode, src_operand, 1, init_op, next_op, prefix);
}


/* Return an instruction template that performs the move without
   clobbering registers used in the input. */
const char *
msp430_mov_noclobber (rtx operands[])
{
  static const char si_fwdtempl[] = "mov\t%A1, %A0\n\tmov\t%B1, %B0";
  static const char si_revtempl[] = "mov\t%B1, %B0\n\tmov\t%A1, %A0";
  static const char di_fwdtempl[] =
    "mov\t%A1, %A0\n\tmov\t%B1, %B0\n\tmov\t%C1, %C0\n\tmov\t%D1, %D0";
  static const char di_revtempl[] =
    "mov\t%D1, %D0\n\tmov\t%C1, %C0\n\tmov\t%B1, %B0\n\tmov\t%A1, %A0";
  static const char di_xBtempl[] =
    "mov\t%A1, %A0\n\tmov\t%C1, %C0\n\tmov\t%D1, %D0\n\tmov\t%B1, %B0";
  static const char di_xCtempl[] =
    "mov\t%A1, %A0\n\tmov\t%B1, %B0\n\tmov\t%D1, %D0\n\tmov\t%C1, %C0";

  const char *fwdtempl;
  const char *revtempl;
  int dst_regno;
  int dst_end_regno;
  int index_reg;

  switch (GET_MODE (operands[0]))
    {
    case SFmode:
    case SImode:
      fwdtempl = si_fwdtempl;
      revtempl = si_revtempl;
      break;
    case DImode:
      fwdtempl = di_fwdtempl;
      revtempl = di_revtempl;
      break;
    default:
      debug_rtx (operands[0]);
      gcc_unreachable ();
    }

  if (!REG_P (operands[0])
      || safe_from_earlyclobber (operands[1], operands[0]))
    return fwdtempl;

  if (REG_P (operands[1]))
    return REGNO (operands[0]) > REGNO (operands[1]) ? revtempl : fwdtempl;

  gcc_assert (MEM_P (operands[1]));
  dst_regno = REGNO (operands[0]);
  dst_end_regno = END_REGNO (operands[0]);
  for (index_reg = dst_regno; index_reg < dst_end_regno; ++index_reg)
    if (reg_mentioned_p (gen_rtx_REG (VOIDmode, index_reg), operands[1]))
      break;
  if (index_reg == dst_regno)
    return revtempl;
  if (index_reg + 1 == dst_end_regno)
    return fwdtempl;
  if (index_reg == dst_regno + 1)
    return di_xBtempl;
  if (index_reg == dst_regno + 2)
    return di_xCtempl;

  gcc_unreachable ();
}

/* Conservative guess whether jump is going to be in range.  If false
   positive, assembler or linker will reject.  If false negative, code
   will be larger than necessary.  Err on the side of working code.
   
   Note that the distance is in units (bytes), not words; the
   underlying instruction supports -1024 to 1022 as it stores the
   offset in words as a signed 10-bit value. */
#define MSP430_JUMP_IN_RANGE(_d) (-500 <= (_d) && (_d) <= 500)

/* Offset from insn to dest in bytes.  If addresses are not known,
   returns a value known to be considered "out of range" for a
   conditional jump. */
int
msp430_jump_dist (rtx dest, rtx insn)
{
  int dest_addr;
  int cur_addr;

  if (!insn_addresses_)
    return 1026;

  dest_addr = INSN_ADDRESSES (INSN_UID (GET_CODE (dest) == LABEL_REF
					? XEXP (dest, 0) : dest));
  cur_addr = INSN_ADDRESSES (INSN_UID (insn));
  return dest_addr - cur_addr;
}


static rtx
msp430_libcall_value (enum machine_mode mode, const_rtx fun ATTRIBUTE_UNUSED)
{
  int words;

  words = (GET_MODE_SIZE (mode) + UNITS_PER_WORD - 1) / UNITS_PER_WORD;
  if (words > RETURN_REGISTERS_AVAILABLE)
    internal_error ("Too many words (%d) for libcall register return", words);
  return gen_rtx_REG (mode, (MSP430_RETURN_REGISTER_BASE + 1 - words));
}

static rtx
msp430_function_value (const_tree type,
		       const_tree fn_decl_or_type ATTRIBUTE_UNUSED,
		       bool outgoing ATTRIBUTE_UNUSED)
{
  unsigned HOST_WIDE_INT bytes;
  int words;
  enum machine_mode tmpmode;

  if (TYPE_MODE (type) != BLKmode)
    return msp430_libcall_value (TYPE_MODE (type), 0);

  /* Although we return reg:BLK, explow.c:hard_function_value will
     override this to be an integral mode, and we have to round the
     size up so the return register is compatible. */
  bytes = int_size_in_bytes (type);
  tmpmode = GET_CLASS_NARROWEST_MODE (MODE_INT);
  while ((tmpmode != VOIDmode) && (GET_MODE_SIZE (tmpmode) < bytes))
    tmpmode = GET_MODE_WIDER_MODE (tmpmode);
  gcc_assert (tmpmode != VOIDmode);
  words = (GET_MODE_SIZE (tmpmode) + UNITS_PER_WORD - 1) / UNITS_PER_WORD;
  if (words > RETURN_REGISTERS_AVAILABLE)
    internal_error ("Too many words (%d) for block register return", words);

  return gen_rtx_REG (BLKmode, (MSP430_RETURN_REGISTER_BASE + 1 - words));
}

/* 1 if REGNO is a general register or the stack pointer */
int
msp430_general_or_stack_reg (int regno)
{
  return MSP430_GENERAL_REGISTER_NUM_P (regno)
    || STACK_POINTER_REGNUM == regno;
}

/* 1 if X is a memory reference through a register suitable for
   indirect referencing: @rN. */
int
msp430_indirect_register_operand (rtx x)
{
  if (!MEM_P (x))
    return 0;
  x = XEXP (x, 0);
  if (!REG_P (x))
    return 0;
  return msp430_general_or_stack_reg (REGNO (x));
}

/* 1 if X is a indexed memory reference: X(rN). */
int
msp430_indexed_register_operand (rtx x)
{
  if (!MEM_P (x))
    return 0;
  x = XEXP (x, 0);
  if (GET_CODE (x) != PLUS)
    return 0;
  /* Left addend (Rn) must be an indexing register, but assume that
     the assembler will be able to create a constant out of whatever
     the right addend (X) is. */
  return msp430_general_or_stack_reg (REGNO (XEXP (x, 0)));
}

/* The routine used to output NUL terminated strings.  We use a special
version of this for most svr4 targets because doing so makes the
generated assembly code more compact (and thus faster to assemble)
as well as more readable, especially for targets like the i386
(where the only alternative is to output character sequences as
comma separated lists of numbers).   */

#define ESCAPES \
"\1\1\1\1\1\1\1\1btn\1fr\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\
\0\0\"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\
\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\\\0\0\0\
\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\1\
\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\
\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\
\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\
\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1\1"
/* A table of bytes codes used by the ASM_OUTPUT_ASCII and
   ASM_OUTPUT_LIMITED_STRING macros.  Each byte in the table
   corresponds to a particular byte value [0..255].  For any
   given byte value, if the value in the corresponding table
   position is zero, the given character can be output directly.
   If the table value is 1, the byte must be output as a \ooo
   octal escape.  If the tables value is anything else, then the
   byte value should be output as a \ followed by the value
   in the table.  Note that we can use standard UN*X escape
   sequences for many control characters, but we don't use
   \a to represent BEL because some svr4 assemblers (e.g. on
   the i386) don't know about that.  Also, we don't use \v
   since some versions of gas, such as 2.2 did not accept it.  */

void
gas_output_limited_string (FILE * file, const char *str)
{
  const unsigned char *_limited_str = (const unsigned char *) str;
  unsigned ch;
  fputs ("\t.string\t\"", file);
  for (; (ch = *_limited_str); _limited_str++)
    {
      int escape;
      switch (escape = ESCAPES[ch])
	{
	case 0:
	  putc (ch, file);
	  break;
	case 1:
	  fprintf (file, "\\%03o", ch);
	  break;
	default:
	  putc ('\\', file);
	  putc (escape, file);
	  break;
	}
    }
  fprintf (file, "\"\n");
}

/* The routine used to output sequences of byte values.  We use a special
version of this for most svr4 targets because doing so makes the
generated assembly code more compact (and thus faster to assemble)
as well as more readable.  Note that if we find subparts of the
character sequence which end with NUL (and which are shorter than
STRING_LIMIT) we output those using ASM_OUTPUT_LIMITED_STRING.  */

#define STRING_LIMIT	((unsigned) 64)

void
gas_output_ascii (FILE * file, const char *str, size_t length)
{
  const unsigned char *_ascii_bytes = (const unsigned char *) str;
  const unsigned char *limit = _ascii_bytes + length;
  unsigned bytes_in_chunk = 0;
  for (; _ascii_bytes < limit; _ascii_bytes++)
    {
      const unsigned char *p;
      if (bytes_in_chunk >= 60)
	{
	  fprintf (file, "\"\n");
	  bytes_in_chunk = 0;
	}
      for (p = _ascii_bytes; p < limit && *p != '\0'; p++)
	continue;
      if (p < limit && (p - _ascii_bytes) <= (signed) STRING_LIMIT)
	{
	  if (bytes_in_chunk > 0)
	    {
	      fprintf (file, "\"\n");
	      bytes_in_chunk = 0;
	    }
	  gas_output_limited_string (file, (const char *) _ascii_bytes);
	  _ascii_bytes = p;
	}
      else
	{
	  int escape;
	  unsigned ch;
	  if (bytes_in_chunk == 0)
	    fprintf (file, "\t.ascii\t\"");
	  switch (escape = ESCAPES[ch = *_ascii_bytes])
	    {
	    case 0:
	      putc (ch, file);
	      bytes_in_chunk++;
	      break;
	    case 1:
	      fprintf (file, "\\%03o", ch);
	      bytes_in_chunk += 4;
	      break;
	    default:
	      putc ('\\', file);
	      putc (escape, file);
	      bytes_in_chunk += 2;
	      break;
	    }
	}
    }
  if (bytes_in_chunk > 0)
    fprintf (file, "\"\n");
}


static section *
msp430_select_section (tree decl, int reloc, unsigned HOST_WIDE_INT align)
{
  /* For now, the only thing we override is const strings.  In the
   * future we might be looking at near/far attributes. */
  if (TREE_CODE (decl) == STRING_CST)
    return default_function_rodata_section (current_function_decl);
  return default_select_section (decl, reloc, align);
}

/* Outputs to the stdio stream FILE some
appropriate text to go at the start of an assembler file.  */

static void
msp430_file_start (void)
{
  FILE *file = asm_out_file;
  const struct tag_value_pair_t *mp;

  output_file_directive (file, main_input_filename);
  if (msp430_mcu_name)
    fprintf (file, "\t.arch %s\n", msp430_mcu_name);
  mp = find_pair_by_value (msp430_cpu, cpu_tag_value_map);
  if (mp)
    fprintf (file, "\t.cpu %s\n", mp->tag);
  mp = find_pair_by_value (msp430_mpy, mpy_tag_value_map);
  if (mp)
    fprintf (file, "\t.mpy %s\n", mp->tag);
  fprintf (file, "\n");
}

int
msp430_hard_regno_mode_ok (int regno ATTRIBUTE_UNUSED,
			   enum machine_mode mode ATTRIBUTE_UNUSED)
{
  return 1;
}

static bool
msp430_frame_pointer_required (void)
{
  return cfun->machine->frame_pointer_required;
}

static bool
msp430_can_eliminate (const int from_reg ATTRIBUTE_UNUSED, const int to_reg)
{
  if (frame_pointer_needed)
    return to_reg == HARD_FRAME_POINTER_REGNUM;
  return true;
}

/* Adjustment to overall insn length if the given integer value is the
   source for an operation in the given value.  Assumption is that the
   operation consists of a single MSP430 instruction repeated once for
   each word in mode. */
static int
length_adjustment_for_int (HOST_WIDE_INT ival, enum machine_mode mode)
{
  HOST_WIDE_INT i;
  int adjustment = 0;

  switch (mode)
    {
    case DImode:
#if HOST_BITS_PER_WIDE_INT > 32
      i = trunc_int_for_mode (0xFFFF & (ival >> 48), HImode);
      if (MSP430_CG_INT_P (i))
	adjustment -= UNITS_PER_WORD;
      i = trunc_int_for_mode (0xFFFF & (ival >> 32), HImode);
      if (MSP430_CG_INT_P (i))
	adjustment -= UNITS_PER_WORD;
#endif /* HOST_BITS_PER_WIDE_INT */
      /* FALLTHRU */
    case SImode:
    case SFmode:
      i = trunc_int_for_mode (0xFFFF & (ival >> 16), HImode);
      if (MSP430_CG_INT_P (i))
	adjustment -= UNITS_PER_WORD;
      /* FALLTHRU */
    case HImode:
    case QImode:
      i = trunc_int_for_mode (0xFFFF & (ival >> 0), HImode);
      if (MSP430_CG_INT_P (i))
	adjustment -= UNITS_PER_WORD;
      break;
    default:
      break;
    }
  return adjustment;
}

/* Adjustment to instruction length if the given op is an immediate
   value source for an operation in the given mode. */
static int
length_adjustment_for_immediate (rtx op, enum machine_mode mode)
{
  int adjustment = 0;
  HOST_WIDE_INT ival;

  switch (GET_CODE (op))
    {
    case CONST_INT:
      ival = INTVAL (op);
#if HOST_BITS_PER_WIDE_INT <= 32
      if (DImode == mode)
	adjustment -= 2 * UNITS_PER_WORD;
#endif /* HOST_BITS_PER_WIDE_INT */
      break;
    case CONST_DOUBLE:
#if HOST_BITS_PER_WIDE_INT <= 32
      if (4 < GET_MODE_SIZE (mode))
	adjustment +=
	  length_adjustment_for_int (CONST_DOUBLE_HIGH (op), mode);
#endif /* HOST_BITS_PER_WIDE_INT */
      ival = CONST_DOUBLE_LOW (op);
      break;
    case SYMBOL_REF:
    case LABEL_REF:
    case CONST:
      return 0;
    default:
      gcc_unreachable ();
    }
  adjustment += length_adjustment_for_int (ival, mode);
  return adjustment;
}

/* Return the length adjusted for the effects of pattern in the
   context of insn. */
static int
adjust_pattern_length (rtx insn, rtx pattern, int len)
{
#if TRACE_INSN_LENGTH
  int ilen = len;
#endif

  switch (GET_CODE (pattern))
    {
    case SET:
      {
	bool skip_adjustment = false;
	enum rtx_code src_code;
	bool output_uses_S;
	rtx src = SET_SRC (pattern);
	rtx src0;
	rtx dst = SET_DEST (pattern);
	enum machine_mode mode = GET_MODE (dst);
	enum attr_instr_format format = get_attr_instr_format (insn);

	if (INSTR_FORMAT_UNDEF == format)
	  break;

	output_uses_S = push_operand (dst, mode);
	src_code = GET_CODE (src);
	switch (src_code)
	  {
	  case UNSPEC:
	    switch (XINT (src, 1))
	      {
	      case UNSPEC_PUSH_MULTI:
	      case UNSPEC_POP_MULTI:
		src = gen_rtx_REG (HImode, INTVAL (XVECEXP (src, 0, 1)));
		break;
	      default:
#if CHECK_INSN_LENGTH
		fprintf (stderr, "Unrecognized unspec code %d in source",
			 XINT (src, 1));
		debug_rtx (src);
#endif /* CHECK_INSN_LENGTH */
		break;
	      case UNSPEC_BITTEST_FOR_CARRY:
		{
		  gcc_assert (dst == cc0_rtx || CCmode == GET_MODE (dst));
		  dst = XVECEXP (src, 0, 0);
		  src = XVECEXP (src, 0, 1);
		  break;
		}
	      }
	    break;
	  case UNSPEC_VOLATILE:
	    switch (XINT (src, 1))
	      {
	      case UNSPECV_GET_WATCHDOG_CLEAR_VALUE:
		len += UNITS_PER_WORD;
		src = XVECEXP (src, 0, 0);
		break;
	      default:
		src = XVECEXP (src, 0, 0);
		break;
	      }
	    break;
	  case NOT:
	  case SIGN_EXTEND:
	  case ZERO_EXTEND:
	  case BSWAP:
	    src = XEXP (src, 0);
	    break;
	  case AND:		/* includes NAND */
	  case IOR:
	  case XOR:
	  case PLUS:
	  case MINUS:
	  case ASHIFT:
	    src0 = XEXP (src, 0);
	    if (GET_CODE (src0) == UNSPEC_VOLATILE)
	      src0 = XVECEXP (src0, 0, 0);
	    if (rtx_equal_p (dst, src0))
	      src = XEXP (src, 1);
	    else
	      {
		gcc_assert (rtx_equal_p (dst, XEXP (src, 1)));
		src = src0;
	      }
	    if (src_code == AND && GET_CODE (src) == NOT)
	      src = XEXP (src, 0);
	    break;
	  case CALL:
	    src = XEXP (src, 0);
	    gcc_assert (MEM == GET_CODE (src));
	    src = XEXP (src, 0);
	    output_uses_S = true;
	    break;
	  case COMPARE:
	    gcc_assert (dst == cc0_rtx);
	    src0 = XEXP (src, 0);
	    if (GET_CODE (src0) == AND)
	      {
		gcc_assert (rtx_equal_p (XEXP (src, 1), const0_rtx));
		src = src0;
	      }
	    dst = XEXP (src, 0);
	    src = XEXP (src, 1);
	    break;
	  case IF_THEN_ELSE:
	    {
	      static const int condjmp_insn_len = UNITS_PER_WORD;
	      static const int jmp_insn_len = UNITS_PER_WORD;
	      static const int offset_len = UNITS_PER_WORD; /* @todo A20 */
	      rtx cmp = XEXP (src, 0);
	      rtx if_loc = XEXP (src, 1);
	      rtx else_loc = XEXP (src, 2);
	      int dist = msp430_jump_dist (if_loc, insn);
	      bool in_range = MSP430_JUMP_IN_RANGE (dist);

	      gcc_assert (XEXP (cmp, 0) == cc0_rtx);
	      gcc_assert (XEXP (cmp, 1) == const0_rtx);
	      gcc_assert (else_loc == pc_rtx);
	      switch (GET_CODE (cmp))
		{
		case LE:
		case GT:
		case LEU:
		case GTU:
		  /* Simulate unsupported comparisons with an extra local jump */
		  len += condjmp_insn_len;
		  /* FALLTHRU */
		case EQ:
		case NE:
		case LT:
		case GE:
		case LTU:
		case GEU:
		  /* Out of range will jump around an added br insn
		     plus immediate offset (@todo A20 update) */
		  if (! in_range)
		    len += jmp_insn_len + offset_len;
		  break;
		default:
		  gcc_unreachable ();
		}
	      skip_adjustment = true;
	    }
	    break;
	  case RETURN:
	    skip_adjustment = true;
	    break;
	  case CONST_INT:
	  case MEM:
	  case REG:
	    break;
	  default:
#if CHECK_INSN_LENGTH
	    fprintf (stderr, "Unrecognized source code %d (%s) in pattern ",
		     src_code, rtx_name[src_code]);
	    debug_rtx (pattern);
#endif /* CHECK_INSN_LENGTH */
	    break;
	  }
	if (skip_adjustment)
	  break;
	
	mode = GET_MODE (dst);

	switch (format)
	  {
	  case INSTR_FORMAT_EMU1DD:
	    src = dst;
	    format = INSTR_FORMAT_FMT1;
	    break;
	  case INSTR_FORMAT_FMT1:
	  case INSTR_FORMAT_FMT2:
	  case INSTR_FORMAT_FMT2S:
	  case INSTR_FORMAT_CONDJMP:
	    break;
	  default:
	    gcc_unreachable ();
	  }


	if (immediate_operand (src, VOIDmode))
	  len += length_adjustment_for_immediate (src, mode);
	else if (register_operand (src, VOIDmode))
	  len -= UNITS_PER_WORD * get_attr_instr_mult (insn);
	else if (format == INSTR_FORMAT_FMT1 || format == INSTR_FORMAT_FMT2S)
	  {
	    if (pop_operand (src, VOIDmode))	/* or other auto-increment indirect */
	      len -= UNITS_PER_WORD * get_attr_instr_mult (insn);
	    else if (msp430_indirect_register_operand (src))
	      {
		if (!(format == INSTR_FORMAT_FMT2S
		      && output_uses_S
		      && REGNO (XEXP (src, 0)) == STACK_POINTER_REGNUM))
		  len -= UNITS_PER_WORD;
	      }
	  }
	else
	  {
	    /* indexed and symbolic accounted for in default length. */
#if CHECK_INSN_LENGTH
	    fprintf (stderr, "Unhandled src ");
	    debug_rtx (src);
#endif /* CHECK_INSN_LENGTH */
	  }
	switch (format)
	  {
	  case INSTR_FORMAT_FMT1:
	    if (register_operand (dst, mode) || PC == GET_CODE (dst))
	      len -= UNITS_PER_WORD * get_attr_instr_mult (insn);
	    break;
	  case INSTR_FORMAT_FMT2:
	  case INSTR_FORMAT_FMT2S:
	  case INSTR_FORMAT_CONDJMP:
	    break;
	  default:
	    gcc_unreachable ();
	  }
	break;
      }
    case PARALLEL:
    case SEQUENCE:
      {
	int i;
	for (i = XVECLEN (pattern, 0) - 1; i >= 0; i--)
	  len = adjust_pattern_length (insn, XVECEXP (pattern, 0, i), len);
	break;
      }
    case USE:
      break;
    case CALL:
      {
	rtx src = XEXP (pattern, 0);
	gcc_assert (MEM == GET_CODE (src));
	src = XEXP (src, 0);
	if (immediate_operand (src, VOIDmode))
	  len += length_adjustment_for_immediate (src, HImode);
	else if (register_operand (src, VOIDmode))
	  len -= UNITS_PER_WORD;
	else if (msp430_indirect_register_operand (src))
	  {
	    if (REGNO (XEXP (src, 0)) != STACK_POINTER_REGNUM)
	      len -= UNITS_PER_WORD;
	  }
	break;
      }
    case RETURN:
    case UNSPEC_VOLATILE:
    case CLOBBER:
    case ASM_INPUT:
    case ASM_OPERANDS:
    case ADDR_VEC:
    default:
#if CHECK_INSN_LENGTH
      fprintf (stderr, "Unhandled insn code %d %s\n", GET_CODE (pattern),
	       rtx_name[GET_CODE (pattern)]);
#endif /* CHECK_INSN_LENGTH */
      break;
    }

#if TRACE_INSN_LENGTH
  fprintf (stderr, "insn len %d to %d format %d ", ilen, len,
	   get_attr_instr_format (insn));
  debug_rtx (insn);
#endif /* TRACE_INSN_LENGTH */
  return len;
}

int
msp430_adjust_insn_length (rtx insn, int len)
{
  return adjust_pattern_length (insn, PATTERN (insn), len);
}

/* Return 1 if X and Y are the same for a sign or zero extension,
   i.e. register or memory reference match except in mode. */
int
msp430_extend_matches (const_rtx x, const_rtx y)
{
  int rc = rtx_equal_p (x, y);
  if (rc)
    return rc;
  if (GET_CODE (x) == REG && GET_CODE (y) == SUBREG && SUBREG_BYTE (y) == 0)
    return rtx_equal_p (x, SUBREG_REG (y));
  if (GET_CODE (x) != GET_CODE (y))
    return 0;
  if (GET_CODE (x) == REG)
    return REGNO (x) == REGNO (y);
  if (GET_CODE (x) == MEM)
    return (mpys_rtx == x && mpys_b_rtx == y)
      || (op2_rtx == x && op2_b_rtx == y)
      || rtx_equal_p (XEXP (x, 0), XEXP (y, 0));
  if (GET_CODE (x) == SUBREG)
    return SUBREG_BYTE (x) == SUBREG_BYTE (y)
      && rtx_equal_p (SUBREG_REG (x), SUBREG_REG (y));
  return 0;
}

/* x is the destination of an extend operation where the source is of
   the given mode.  Obtain a compatible reference to the low part of
   the destination so we can copy the source into it. */
static rtx
msp430_extend_force_to_mode (rtx x, enum machine_mode mode)
{
  switch (GET_CODE (x))
    {
    case REG:
      return gen_lowpart (mode, x);
    case MEM:
      return adjust_address (x, mode, 0);
      break;
    case SUBREG:
      return simplify_rtx (gen_lowpart_SUBREG (mode, x));
    default:
      break;
    }
  gcc_unreachable ();
}

/* Expand all {,zero_}extendmn2 instructions. */
static void
msp430_expand_extend (rtx operands[], bool signed_p)
{
  rtx dst_lo = NULL_RTX;	/* hi, or lowpart (hi, si) */
  rtx dst_hi = NULL_RTX;	/* highpart (hi, si) or highpart (hi, lowpart (si, di)) */
  rtx dst_llo = NULL_RTX;	/* lowpart (hi, lowpart (si, di)) */
  rtx dst_lhi = NULL_RTX;	/* highpart (hi, lowpart (si, di)) */
  rtx dst_hlo = NULL_RTX;	/* lowpart (hi, highpart (si, di)) */
  rtx dst_hhi = NULL_RTX;	/* highpart (hi, highpart (si, di)) */
  rtx extended_value;
  bool need_copy;

  switch (GET_MODE (operands[0]))
    {
    case HImode:
      dst_lo = operands[0];
      break;
    case SImode:
      dst_lo = gen_lowpart (HImode, operands[0]);
      dst_hi = gen_highpart (HImode, operands[0]);
      break;
    case DImode:
      if (REG_P (operands[0]))
	{
	  dst_llo = gen_rtx_SUBREG (HImode, operands[0], 0);
	  dst_lhi = gen_rtx_SUBREG (HImode, operands[0], UNITS_PER_WORD);
	  dst_hlo = gen_rtx_SUBREG (HImode, operands[0], 2 * UNITS_PER_WORD);
	  dst_hhi = gen_rtx_SUBREG (HImode, operands[0], 3 * UNITS_PER_WORD);
	}
      else if (MEM_P (operands[0]))
	{
	  dst_llo = adjust_address (operands[0], HImode, 0);
	  dst_lhi = adjust_address (dst_llo, HImode, UNITS_PER_WORD);
	  dst_hlo = adjust_address (dst_lhi, HImode, UNITS_PER_WORD);
	  dst_hhi = adjust_address (dst_hlo, HImode, UNITS_PER_WORD);
	}
      else
	{
	  debug_rtx (operands[0]);
	  gcc_unreachable ();
	}
      dst_lo = dst_llo;
      dst_hi = dst_lhi;
      break;
    default:
      gcc_unreachable ();
    }
  extended_value = signed_p ? NULL_RTX : GEN_INT (0);
  need_copy = !msp430_extend_matches (operands[0], operands[1]);
  if (QImode == GET_MODE (operands[1]))
    {
      rtx dst_qi = msp430_extend_force_to_mode (dst_lo, QImode);

      if (!signed_p && GET_MODE (operands[0]) == HImode
	  && operands[0] == op2_rtx)
	{
	  emit_insn (gen_andhi3 (dst_lo, dst_lo, GEN_INT (0xff00)));
	  return;
	}
      if (need_copy)
	{
	  if (!signed_p && register_operand (dst_lo, HImode))
	    emit_insn (gen_loadqi2 (dst_lo, operands[1]));
	  else
	    emit_move_insn (dst_qi, operands[1]);
	}
      if (signed_p)
	{
	  emit_insn (gen_extend8bithi1 (dst_lo));
	  if (dst_hi)
	    {
	      extended_value = copy_to_reg (dst_lo);
	      emit_insn (gen_bswaphi1 (extended_value));
	      emit_insn (gen_extend8bithi1 (extended_value));
	    }
	}
      else
	{
	  if (memory_operand (operands[0], GET_MODE (operands[0])))
	    emit_move_insn (gen_highpart (QImode, dst_lo), extended_value);
	  else if (!need_copy)
	    emit_insn (gen_andhi3 (dst_lo, dst_lo, GEN_INT (0xff)));
	}
    }
  else if (HImode == GET_MODE (operands[1]))
    {
      if (need_copy)
	emit_move_insn (dst_lo, operands[1]);
      if (signed_p)
	{
	  extended_value = gen_reg_rtx (HImode);
	  emit_insn (gen_ashrhi3 (extended_value, dst_lo, 
				  GEN_INT (GET_MODE_BITSIZE (HImode) - 1)));
	}
    }
  else if (SImode == GET_MODE (operands[1]))
    {
      if (need_copy)
	emit_move_insn (gen_lowpart (SImode, operands[0]), operands[1]);
      if (signed_p)
	{
	  extended_value = gen_reg_rtx (HImode);
	  emit_insn (gen_ashrhi3 (extended_value, dst_hi, 
				  GEN_INT (GET_MODE_BITSIZE (HImode) - 1)));
	}
      dst_hi = NULL_RTX;
    }
  else
    gcc_unreachable ();
  if (dst_hi)
    emit_move_insn (dst_hi, extended_value);
  if (dst_hlo)
    {
      emit_move_insn (dst_hlo, extended_value);
      emit_move_insn (dst_hhi, extended_value);
    }
}

/***** SIGN EXTEND *********/

/* Expand all (signed) extendmn2 instructions.  Return 1 if all
   necessary instructions have been emited; returns 0 to fall through
   and generate extendqihi2_match. */
void
msp430_expand_signextend (rtx operands[])
{
  msp430_expand_extend (operands, true);
}

/**** ZERO EXTEND *****/

/* Expand all zero_extendmn2 instructions. */
void
msp430_expand_zeroextend (rtx operands[])
{
  msp430_expand_extend (operands, false);
}

int
msp430_expand_extract (rtx operands[], bool signed_p)
{
  rtx dst = operands[0];
  rtx bits = operands[1];
  rtx width = operands[2];
  rtx start = operands[3];
  HOST_WIDE_INT mask;

  if (!CONST_INT_P (width) || !CONST_INT_P (start))
    return 0;
  if (1 != INTVAL (width))
    return 0;
  if (signed_p)
    return 0;
  mask = (HOST_WIDE_INT)1 << INTVAL (start);
  if (trunc_int_for_mode (mask, GET_MODE (bits)) != mask)
    return 0;
  gcc_assert (HImode == GET_MODE (dst));
  if (GET_MODE (bits) == QImode)
    emit_insn (gen_bittestforcarryqi2 (bits, GEN_INT (mask)));
  else
    {
      gcc_assert (HImode == GET_MODE (bits));
      emit_insn (gen_bittestforcarryhi2 (bits, GEN_INT (mask)));
    }
  emit_move_insn (dst, const0_rtx);
  emit_insn (gen_rlchi1 (dst));
  return 1;
}

static int
mpy_for_hardware_multiply (enum machine_mode result_mode,
			   enum machine_mode operand_mode)
{
  bool widen;
  int required_mpy = 0;

  if (VOIDmode == operand_mode)
    operand_mode = result_mode;
  widen = (operand_mode != result_mode);
  switch (result_mode)
    {
    case QImode:
      required_mpy = MSP430_MPY_TYPE_16;
      break;
    case HImode:
      required_mpy = MSP430_MPY_TYPE_16;
      break;
    case SImode:
      required_mpy = widen ? MSP430_MPY_TYPE_16 : MSP430_MPY_TYPE_32;
      break;
    case DImode:
      required_mpy = widen ? MSP430_MPY_TYPE_32 : MSP430_MPY_NONE;
      break;
    default:
      required_mpy = MSP430_MPY_NONE;
      break;
    }
  return required_mpy;
}

void
msp430_expand_mul (rtx operands[], int signed_mul)
{
  rtx m_mpy;
  rtx m_op2;
  rtx m_reslo;
  int required_mpy;
  bool widen;
  enum machine_mode op0mode = GET_MODE (operands[0]);
  enum machine_mode op1mode = GET_MODE (operands[1]);
  enum machine_mode op2mode = GET_MODE (operands[2]);

  gcc_assert ((op1mode == op2mode) || CONST_INT_P (operands[2]));
  op2mode = op1mode;
  widen = (op0mode == GET_MODE_WIDER_MODE (op1mode));
  gcc_assert (widen || op0mode == op1mode);

  required_mpy = mpy_for_hardware_multiply (op0mode, op1mode);

  if (!(msp430_mpy & required_mpy))
    {
      rtx libsym = NULL_RTX;

      switch (op0mode)
	{
	case QImode:
	  libsym = libsym_mulqi3;
	  break;
	case HImode:
	  if (widen)
	    libsym = (signed_mul ? libsym_mulqihi3 : libsym_umulqihi3);
	  else
	    libsym = libsym_mulhi3;
	  break;
	case SImode:
	  if (widen)
	    libsym = (signed_mul ? libsym_mulhisi3 : libsym_umulhisi3);
	  else
	    libsym = libsym_mulsi3;
	  break;
	case DImode:
	  if (widen)
	    libsym = (signed_mul ? libsym_mulsidi3 : libsym_umulsidi3);
	  else
	    libsym = libsym_muldi3;
	  break;
	default:
	  gcc_unreachable ();
	}

      emit_library_call_value (libsym, operands[0], LCT_CONST,
			       op0mode, 2,
			       operands[1], op1mode,
			       operands[2], op1mode);
      return;
    }

  if (!MSP430_NOINT_HWMUL)
    emit_insn (gen_mpy_inhibit_intr ());

  if (MSP430_MPY_TYPE_16 == required_mpy)
    {
      m_mpy = signed_mul ? mpys_rtx : mpy_rtx;
      m_op2 = op2_rtx;
      m_reslo = reslo_rtx;
      if (op0mode == QImode)
	m_reslo = reslo_b_rtx;
      if (op1mode == QImode)
	m_mpy = signed_mul ? mpys_b_rtx : mpy_b_rtx;
      if (op2mode == QImode)
	m_op2 = op2_b_rtx;

      emit_move_insn (m_mpy, operands[1]);
      if (op1mode == QImode && signed_mul
	  && !(msp430_mpy & MSP430_MPY_HAS_SE))
	{
	  gcc_assert (m_mpy == mpys_b_rtx);
	  emit_insn (gen_extendqihi2 (mpys_rtx, m_mpy));
	}
      emit_move_insn (m_op2, operands[2]);
      if (op2mode == QImode && signed_mul
	  && !(msp430_mpy & MSP430_MPY_HAS_SE))
	emit_insn (gen_extendqihi2 (op2_rtx, m_op2));

      switch (op0mode)
	{
	case QImode:
	case HImode:
	  emit_move_insn (operands[0], m_reslo);
	  break;
	case SImode:
	  emit_move_insn (gen_lowpart (HImode, operands[0]), reslo_rtx);
	  emit_move_insn (gen_highpart (HImode, operands[0]), reshi_rtx);
	  break;
	default:
	  gcc_unreachable ();
	}
    }
  else
    {
      gcc_assert (SImode == op1mode);
      emit_move_insn (signed_mul ? mpys32l_rtx : mpy32l_rtx,
		      gen_lowpart (HImode, operands[1]));
      emit_move_insn (signed_mul ? mpys32h_rtx : mpy32h_rtx,
		      gen_highpart (HImode, operands[1]));
      if (CONST_INT_P (operands[2]))
	{
	  HOST_WIDE_INT iv = INTVAL (operands[2]);
	  emit_move_insn (op2l_rtx,
			  GEN_INT (trunc_int_for_mode
				   (iv & ((1 << BITS_PER_WORD) - 1),
				    HImode)));
	  emit_move_insn (op2h_rtx,
			  GEN_INT (trunc_int_for_mode
				   (iv >> BITS_PER_WORD, HImode)));
	}
      else
	{
	  gcc_assert (SImode == op2mode);
	  emit_move_insn (op2l_rtx, gen_lowpart (HImode, operands[2]));
	  emit_move_insn (op2h_rtx, gen_highpart (HImode, operands[2]));
	}
      switch (op0mode)
	{
	case SImode:
	  emit_move_insn (gen_lowpart (HImode, operands[0]), reslo_rtx);
	  emit_move_insn (gen_highpart (HImode, operands[0]), reshi_rtx);
	  break;
	case DImode:
	  {
	    rtx dst_llo = NULL_RTX;	/* lowpart (hi, lowpart (si, di)) */
	    rtx dst_lhi = NULL_RTX;	/* highpart (hi, lowpart (si, di)) */
	    rtx dst_hlo = NULL_RTX;	/* lowpart (hi, highpart (si, di)) */
	    rtx dst_hhi = NULL_RTX;	/* highpart (hi, highpart (si, di)) */

	    if (REG_P (operands[0]))
	      {
		dst_llo = gen_rtx_SUBREG (HImode, operands[0], 0);
		dst_lhi =
		  gen_rtx_SUBREG (HImode, operands[0], UNITS_PER_WORD);
		dst_hlo =
		  gen_rtx_SUBREG (HImode, operands[0], 2 * UNITS_PER_WORD);
		dst_hhi =
		  gen_rtx_SUBREG (HImode, operands[0], 3 * UNITS_PER_WORD);
	      }
	    else if (MEM_P (operands[0]))
	      {
		dst_llo = adjust_address (operands[0], HImode, 0);
		dst_lhi = adjust_address (dst_llo, HImode, UNITS_PER_WORD);
		dst_hlo = adjust_address (dst_lhi, HImode, UNITS_PER_WORD);
		dst_hhi = adjust_address (dst_hlo, HImode, UNITS_PER_WORD);
	      }
	    else
	      gcc_unreachable ();

	    emit_move_insn (dst_llo, res0_rtx);
	    emit_move_insn (dst_lhi, res1_rtx);
	    emit_move_insn (dst_hlo, res2_rtx);
	    emit_move_insn (dst_hhi, res3_rtx);
	    break;
	  }
	default:
	  gcc_unreachable ();
	}
    }

  if (!MSP430_NOINT_HWMUL)
    emit_insn (gen_mpy_restore_intr ());
}

enum shift_type_e
{
  ST_INVALID,
  ST_ashl,
  ST_ashr,
  ST_lshr,
  ST_lshr_first,
  ST_lshr_next
};

/* Shift types: (arith << 1) | left */

#define ST_IS_ARITH(_st) (ST_ashl == (_st) || ST_ashr == (_st))
#define ST_IS_LEFT(_st) (ST_ashl == (_st))

static void
msp430_expand_libcall_shift (rtx operands[], enum shift_type_e shift_type)
{
  enum machine_mode op0mode = GET_MODE (operands[0]);
  bool arith_shift = ST_IS_ARITH (shift_type);
  bool left_shift = ST_IS_LEFT (shift_type);
  rtx libsym;

  gcc_assert (CONST_INT_P (operands[2]) || QImode == GET_MODE (operands[2]));
  gcc_assert (GET_MODE (operands[1]) == op0mode);
  switch (op0mode)
    {
    case QImode:
      if (arith_shift)
	libsym = (left_shift ? libsym_ashlqi3 : libsym_ashrqi3);
      else
	libsym = libsym_lshrqi3;
      break;
    case HImode:
      if (arith_shift)
	libsym = (left_shift ? libsym_ashlhi3 : libsym_ashrhi3);
      else
	libsym = libsym_lshrhi3;
      break;
    case SImode:
      if (arith_shift)
	libsym = (left_shift ? libsym_ashlsi3 : libsym_ashrsi3);
      else
	libsym = libsym_lshrsi3;
      break;
    case DImode:
      if (arith_shift)
	libsym = (left_shift ? libsym_ashldi3 : libsym_ashrdi3);
      else
	libsym = libsym_lshrdi3;
      break;
    default:
      gcc_unreachable ();
    }
  emit_library_call_value (libsym,
			   operands[0], LCT_CONST,
			   op0mode, 2,
			   operands[1], GET_MODE (operands[0]),
			   operands[2], QImode);
}

static void
expand_shift_bitsizem1 (rtx r_dst, rtx r_src, enum shift_type_e shift_type)
{
  enum machine_mode opmode = GET_MODE (r_dst);
  rtx parts[4];
  rtx r_hi;
  rtx r_fill;
  rtx r_qi = 0;
  int pres;
  int pmax;
  int nparts = msp430_extract_multiword_operand (opmode, r_src, parts);

  /* The only time it's worth considering doing this operation in
   * memory is if src==dst and src is memory and it's an ashl or lshr
   * on QImode or HImode, at which point we save one word of ROM. */
  gcc_assert (0 < nparts);
  pmax = nparts - 1;
  pres = ST_IS_LEFT (shift_type) ? pmax : 0;
  r_hi = gen_reg_rtx (HImode);
  if (QImode == opmode)
    {
      emit_insn (gen_loadqi2 (r_hi, parts[pmax - pres]));
      r_qi = gen_lowpart (QImode, r_hi);
    }
  else
    emit_move_insn (r_hi, parts[pmax - pres]);

  r_fill = const0_rtx;
  if (ST_ashl == shift_type)
    {
      emit_insn (gen_rrahi1 (r_hi));
      emit_move_insn (r_hi, const0_rtx);
      if (QImode == opmode)
	emit_insn (gen_rrcqi1 (r_qi));
      else
	emit_insn (gen_rrchi1 (r_hi));
    }
  else if (ST_ashr == shift_type)
    {
      if (QImode == opmode)
	{
	  emit_insn (gen_extend8bithi1 (r_hi));
	  emit_insn (gen_bswaphi1 (r_hi));
	}
      else
	{
	  emit_insn (gen_bswaphi1 (r_hi));
	  emit_insn (gen_extend8bithi1 (r_hi));
	  emit_insn (gen_bswaphi1 (r_hi));
	  emit_insn (gen_extend8bithi1 (r_hi));
	}
      r_fill = r_hi;
    }
  else
    {
      gcc_assert (ST_lshr == shift_type);
      if (QImode == opmode)
	emit_insn (gen_rlaqi1 (r_qi));
      else
	emit_insn (gen_rlahi1 (r_hi));
      emit_move_insn (r_hi, const0_rtx);
      emit_insn (gen_rlchi1 (r_hi));
    }

  if (QImode == opmode)
    emit_insn (gen_storeqi2 (r_dst, r_hi));
  else
    {
      int pi;

      nparts = msp430_extract_multiword_operand (opmode, r_dst, parts);
      for (pi = 0; pi < nparts; ++pi)
	emit_move_insn (parts[pi], (pres == pi) ? r_hi : r_fill);
    }
}

static void msp430_expand_shift (rtx operands[], enum shift_type_e shift_type);

static void
expand_unary_shift (rtx op, int count, enum shift_type_e shift_type)
{
  rtx operands[3];

  operands[0] = op;
  operands[1] = op;
  operands[2] = GEN_INT (count);
  msp430_expand_shift (operands, shift_type);
}

static const int shift_loop_overhead_words = 5;	/* push?, mov[2?], decr, jne, pop? */

static void
msp430_expand_shift (rtx operands[], enum shift_type_e shift_type)
{
  rtx r_src = operands[1];
  rtx r_dst = operands[0];
  enum machine_mode opmode = GET_MODE (r_dst);
  bool arith_shift = ST_IS_ARITH (shift_type);
  bool left_shift = ST_IS_LEFT (shift_type);
  int count;
  rtx src_parts[4];
  rtx op_parts[4];
  bool op_is_mem;
  bool use_dst_as_op;
  bool need_op_init;
  bool need_op_store;
  int pi;
  int pmin;
  int pmax;
  int nparts;

  if (CONST_INT_P (operands[2]))
    {
      count = INTVAL (operands[2]) % GET_MODE_BITSIZE (opmode);
      operands[2] = GEN_INT (count);
    }
  else
    count = -1;

  gcc_assert (GET_MODE (r_src) == GET_MODE (r_dst));

  /* If effective shift is zero, do nothing */
  if (0 == count)
    {
      if (!rtx_equal_p (r_dst, r_src))
	emit_move_insn (r_dst, r_src);
      return;
    }

  if (count == GET_MODE_BITSIZE (opmode) - 1)
    {
      expand_shift_bitsizem1 (r_dst, r_src, shift_type);
      return;
    }

  if (msp430_enable_libcall_shift)
    {
      bool use_libcall = false;

      /* inline loops are smaller than the overhead of setting up for
       * a library call plus the size of the library routine itself.
       * Assuming that non-constant shifts are rare, prefer libcalls
       * for non-constant shifts unless optimizing for size. */
      if (!optimize_function_for_size_p (cfun) && QImode != opmode
	  && 0 > count)
	use_libcall = true;

      if (use_libcall)
	{
	  msp430_expand_libcall_shift (operands, shift_type);
	  return;
	}
    }

  /* Canonicalize the count, or note that it's determined at runtime */
  op_is_mem = false;
  need_op_store = need_op_init = true;
  if (rtx_equal_p (r_src, r_dst) && !register_operand (r_dst, VOIDmode))
    {
      /* Potentially decide to put op in memory. */
    }

  nparts = msp430_extract_multiword_operand (opmode, r_src, src_parts);
  pmin = 0;
  pmax = nparts - 1;

  memset (op_parts, 0, sizeof (op_parts));

  use_dst_as_op = false;
  if (op_is_mem)
    use_dst_as_op = true;
  else if (QImode == opmode)
    {
      op_parts[pmin] = gen_reg_rtx (HImode);
      emit_insn (gen_loadqi2 (op_parts[pmin], src_parts[pmin]));
      if (ST_ashr == shift_type)
	emit_insn (gen_extend8bithi1 (op_parts[pmin]));
      need_op_init = false;
    }
  else if ((count < BITS_PER_WORD) && register_operand (r_dst, VOIDmode))
    use_dst_as_op = true;

  if (use_dst_as_op)
    {
      if (!rtx_equal_p (r_dst, r_src))
	emit_move_insn (r_dst, r_src);
      msp430_extract_multiword_operand (opmode, r_dst, op_parts);
      need_op_store = need_op_init = false;
    }
  else if (!op_parts[0])
    {
      for (pi = 0; pi < nparts; ++pi)
	op_parts[pi] = gen_reg_rtx (HImode);
    }

  if (count >= 48)
    {
      rtx newop = NULL_RTX;
      gcc_assert (4 == nparts);
      gcc_assert (need_op_init);
      if (ST_IS_LEFT (shift_type))
	{
	  emit_move_insn (op_parts[3], src_parts[0]);
	  emit_move_insn (op_parts[2], const0_rtx);
	  emit_move_insn (op_parts[1], const0_rtx);
	  emit_move_insn (op_parts[0], const0_rtx);
	  newop = op_parts[3];
	}
      else
	{
	  emit_move_insn (op_parts[0], src_parts[3]);
	  if (ST_IS_ARITH (shift_type))
	    emit_insn (gen_ashrhi_15 (op_parts[3], op_parts[0]));
	  else
	    emit_move_insn (op_parts[3], const0_rtx);
	  emit_move_insn (op_parts[2], op_parts[3]);
	  emit_move_insn (op_parts[1], op_parts[3]);
	  newop = op_parts[0];
	}
      need_op_init = false;
      count -= 48;
      if (0 < count)
	{
	  expand_unary_shift (newop, count, shift_type);
	  count = 0;
	}
      goto done_op;
    }

  if (count >= 32)
    {
      int base;
      rtx r_fill = const0_rtx;
      rtx newop = gen_reg_rtx (SImode);

      gcc_assert (4 == nparts);
      if (ST_IS_LEFT (shift_type))
	{
	  emit_move_insn (gen_lowpart (HImode, newop), src_parts[0]);
	  emit_move_insn (gen_highpart (HImode, newop), src_parts[1]);
	  base = 2;
	}
      else
	{
	  emit_move_insn (gen_lowpart (HImode, newop), src_parts[2]);
	  emit_move_insn (gen_highpart (HImode, newop), src_parts[3]);
	  base = 0;
	  if (ST_IS_ARITH (shift_type))
	    {
	      r_fill = gen_reg_rtx (HImode);
	      emit_insn (gen_ashrhi_15 (r_fill, src_parts[3]));
	    }
	}
      count -= 32;
      if (0 < count)
	{
	  expand_unary_shift (newop, count, shift_type);
	  count = 0;
	}
      emit_move_insn (op_parts[base], gen_lowpart (HImode, newop));
      emit_move_insn (op_parts[base + 1], gen_highpart (HImode, newop));
      emit_move_insn (op_parts[2 - base], r_fill);
      emit_move_insn (op_parts[2 - base + 1], r_fill);
      need_op_init = false;
      goto done_op;
    }

  if (count >= 16)
    {
      gcc_assert (need_op_init);
      if (ST_IS_LEFT (shift_type))
	{
	  for (pi = pmax; pmin < pi; --pi)
	    emit_move_insn (op_parts[pi], src_parts[pi - 1]);
	  emit_move_insn (op_parts[pmin], const0_rtx);
	  ++pmin;
	}
      else
	{
	  for (pi = pmin; pi < pmax; ++pi)
	    emit_move_insn (op_parts[pi], src_parts[pi + 1]);
	  if (ST_IS_ARITH (shift_type))
	    emit_insn (gen_ashrhi_15 (op_parts[pmax], src_parts[pmax]));
	  else
	    emit_move_insn (op_parts[pmax], const0_rtx);
	  --pmax;
	}
      need_op_init = false;
      count -= 16;
    }

  if (count >= 8)
    {
      if (need_op_init)
	{
	  for (pi = pmin; pi <= pmax; ++pi)
	    emit_move_insn (op_parts[pi], src_parts[pi]);
	  need_op_init = false;
	}

      if (ST_IS_LEFT (shift_type))
	{
	  for (pi = pmax; pmin < pi; --pi)
	    emit_insn (gen_mwshl8xor (op_parts[pi], op_parts[pi - 1]));
	  emit_insn (gen_loadqi2
		     (op_parts[pmin], gen_lowpart (QImode, op_parts[pmin])));
	  emit_insn (gen_bswaphi1 (op_parts[pmin]));
	}
      else
	{
	  emit_insn (gen_bswaphi1 (op_parts[pmin]));
	  for (pi = pmin; pi < pmax; ++pi)
	    emit_insn (gen_mwshr8xor (op_parts[pi], op_parts[pi + 1]));
	  if (ST_ashr == shift_type)
	    emit_insn (gen_extend8bithi1 (op_parts[pmax]));
	  else
	    emit_insn (gen_loadqi2
		       (op_parts[pmax],
			gen_lowpart (QImode, op_parts[pmax])));
	}
      count -= 8;
    }

  if (need_op_init)
    for (pi = 0; pi < nparts; ++pi)
      emit_move_insn (op_parts[pi], src_parts[pi]);

  if (0 < count && msp430_cpu & MSP430_CPU_MSP430X && !op_is_mem && 1 == nparts)
    {
      rtx (*gen_shift_m) (rtx, rtx, rtx);

      gen_shift_m = left_shift ? gen_ashlhi_m : (arith_shift ? gen_ashrhi_m : gen_lshrhi_m);
      while (count >= MSP430_CPUX_MULTISHIFT_MAX)
	{
	  emit_insn (gen_shift_m
		     (op_parts[pmin], op_parts[pmin],
		      GEN_INT (MSP430_CPUX_MULTISHIFT_MAX)));
	  count -= MSP430_CPUX_MULTISHIFT_MAX;
	}
      if (count)
	{
	  emit_insn (gen_shift_m
		     (op_parts[pmin], op_parts[pmin], GEN_INT (count)));
	  count -= count;
	}
    }

done_op:
  if (need_op_store)
    {
      if (QImode == opmode && HImode == GET_MODE (op_parts[0]))
	emit_insn (gen_storeqi2 (r_dst, op_parts[0]));
      else
	{
	  rtx dst_parts[4];
	  msp430_extract_multiword_operand (opmode, r_dst, dst_parts);
	  for (pi = 0; pi < nparts; ++pi)
	    emit_move_insn (dst_parts[pi], op_parts[pi]);
	}
    }

  if (0 != count)
    {
      rtx (*gen_shift_1) (rtx);
      rtx r_cnt = NULL_RTX;
      rtx l_top = NULL_RTX;
      rtx l_bottom = NULL_RTX;
      rtx t_ne = NULL_RTX;
      rtx t_eq = NULL_RTX;
      bool use_loop;

      switch (opmode)
	{
	case QImode:
	  gen_shift_1 =
	    left_shift ? gen_ashlqi_1 : (arith_shift ? gen_ashrqi_1 :
					 gen_lshrqi_1);
	  break;
	case HImode:
	  gen_shift_1 =
	    left_shift ? gen_ashlhi_1 : (arith_shift ? gen_ashrhi_1 :
					 gen_lshrhi_1);
	  break;
	case SImode:
	  gen_shift_1 =
	    left_shift ? gen_ashlsi_1 : (arith_shift ? gen_ashrsi_1 :
					 gen_lshrsi_1);
	  break;
	case DImode:
	  gen_shift_1 =
	    left_shift ? gen_ashldi_1 : (arith_shift ? gen_ashrdi_1 :
					 gen_lshrdi_1);
	  break;
	default:
	  gcc_unreachable ();
	}

      if (0 > count)
	use_loop = true;
      else
	{
	  int shift_1_words;
	  int inline_words;
	  int loop_words;

	  /* Warning: gcc without -Os will unroll loops up to eight
	   * times, so since we know the count will never exceed 7
	   * this doesn't really do much good. */
	  shift_1_words =
	    (GET_MODE_SIZE (opmode) + UNITS_PER_WORD - 1) / UNITS_PER_WORD;
	  inline_words = count * shift_1_words;
	  loop_words = shift_loop_overhead_words + shift_1_words;
	  use_loop =
	    inline_words >=
	    (optimize_function_for_size_p (cfun) ? 1 : 4) * loop_words;
	}

      if (use_loop)
	{
	  r_cnt = gen_reg_rtx (QImode);
	  l_top = gen_label_rtx ();
	  t_ne = gen_rtx_NE (QImode, r_cnt, const0_rtx);

	  if (0 < count)
	    emit_move_insn (r_cnt, GEN_INT (count));
	  else
	    {
	      l_bottom = gen_label_rtx ();
	      t_eq = gen_rtx_EQ (QImode, r_cnt, const0_rtx);
	      emit_move_insn (r_cnt, operands[2]);
	      emit_insn (gen_andqi3
			 (r_cnt, r_cnt,
			  GEN_INT (GET_MODE_BITSIZE (GET_MODE (r_dst)) - 1)));
	      emit_jump_insn (gen_cbranchqi4
			      (t_eq, r_cnt, const0_rtx, l_bottom));
	    }
	  emit_label (l_top);
	  count = 1;
	}

      while (count--)
	emit_insn (gen_shift_1 (r_dst));

      if (l_top)
	{
	  emit_move_insn (r_cnt, plus_constant (r_cnt, -1));
	  emit_jump_insn (gen_cbranchqi4 (t_ne, r_cnt, const0_rtx, l_top));
	  if (l_bottom)
	    emit_label (l_bottom);
	}
    }

  return;
}

void
msp430_expand_ashl (rtx operands[])
{
  msp430_expand_shift (operands, ST_ashl);
}

void
msp430_expand_ashr (rtx operands[])
{
  msp430_expand_shift (operands, ST_ashr);
}

void
msp430_expand_lshr (rtx operands[])
{
  msp430_expand_shift (operands, ST_lshr);
}

void
msp430_notice_update_cc (rtx exp ATTRIBUTE_UNUSED, rtx insn)
{
  rtx set = single_set (insn);
  rtx src;
  enum attr_cc cc = get_attr_cc (insn);

  switch (cc)
    {
    case CC_UNDEF:
    default:
      gcc_unreachable ();
    case CC_NONE:
      if (set)
	{
	  if (cc_status.value1 && modified_in_p (cc_status.value1, insn))
	    cc_status.value1 = NULL_RTX;
	  if (cc_status.value2 && modified_in_p (cc_status.value2, insn))
	    cc_status.value2 = NULL_RTX;
	}
      return;
    case CC_CLOBBER:
      CC_STATUS_INIT;
      break;
    case CC_N:
    case CC_NZ:
    case CC_NZC:
    case CC_VNZC:
    case CC_EXPLICIT:
      CC_STATUS_INIT;
      gcc_assert (set);
      switch (cc)
	{
	case CC_N:
	  cc_status.flags |= CC_CLOBBER_V | CC_CLOBBER_Z | CC_CLOBBER_C;
	  break;
	case CC_NZ:
	  cc_status.flags |= CC_CLOBBER_V | CC_CLOBBER_C;
	  break;
	case CC_NZC:
	  cc_status.flags |= CC_NO_OVERFLOW;
	  break;
	case CC_EXPLICIT:
	  cc_status.flags |= CC_EXPLICIT_COMPARE;
	  break;
	default:
	  break;
	}

      if (rtx_equal_p (cc0_rtx, SET_DEST (set)))
	{
	  src = SET_SRC (set);
	  gcc_assert (COMPARE == GET_CODE (src));
	  cc_status.value1 = src;
	  if (XEXP (src, 1) == const0_rtx)
	    cc_status.value2 = XEXP (src, 0);
	}
      else
	{
	  cc_status.value1 = SET_DEST (set);
	  if (!side_effects_p (SET_SRC (set)))
	    cc_status.value2 = SET_SRC (set);
	}
      break;
    }
  return;
}

static void
expand_cbranch (enum rtx_code code,
		rtx op0,
		rtx op1,
		rtx loc)
{
  rtx cmp;
  enum machine_mode mode;
  bool ps0;
  bool ps1;

  /* If dest prefers source addressing, and source doesn't, swap
     them */
  ps0 = CONSTANT_P (op0) || (MEM_P (op0) && REG_P (XEXP (op0, 0)));
  ps1 = CONSTANT_P (op1) || (MEM_P (op1) && REG_P (XEXP (op1, 0)));
  if (ps0 && !ps1)
    {
      rtx tmp = op0;
      op0 = op1;
      op1 = tmp;
      code = swap_condition (code);
    }
  if (GT == code || GTU == code || LE == code || LEU == code)
    {
      if (CONSTANT_P (op1))
	{
	  if (GET_CODE (op1) == CONST_INT)
	    {
	      HOST_WIDE_INT new_bound = 1 + INTVAL (op1);
	      HOST_WIDE_INT new_bound_trunc;

	      new_bound_trunc = trunc_int_for_mode (new_bound, GET_MODE (op0));
	      if ((new_bound == new_bound_trunc)
		  && ((0 != new_bound)
		      || (code != GTU && code != LEU)))
		{
		  op1 = GEN_INT (new_bound);
		  switch (code)
		    {
		    case GT:
		      code = GE;
		      break;
		    case LE:
		      code = LT;
		      break;
		    case GTU:
		      code = GEU;
		      break;
		    case LEU:
		      code = LTU;
		      break;
		    default:
		      gcc_unreachable ();
		    }
		}
	    }
	}
      else
	{
	  rtx tmp = op0;
	  op0 = op1;
	  op1 = tmp;
	  code = swap_condition (code);
	}
    }

  mode = GET_MODE (op0);

  gcc_assert (GET_MODE_CLASS (mode) == MODE_INT);
  if (GET_MODE_SIZE (mode) > UNITS_PER_WORD)
    {
      enum rtx_code rcode = reverse_condition (code);
      rtx fail = gen_label_rtx ();
      rtx parts0[4];
      rtx parts1[4];
      int words;
      int w;

      words = msp430_extract_multiword_operand (mode, op0, parts0);
      msp430_extract_multiword_operand (mode, op1, parts1);
      switch (code)
	{
	case EQ:
	  for (w = 0; w < words - 1; ++w)
	    expand_cbranch (rcode, parts0[w], parts1[w], fail);
	  expand_cbranch (code, parts0[w], parts1[w], loc);
	  emit_label (fail);
	  break;
	case NE:
	  for (w = 0; w < words; ++w)
	    expand_cbranch (code, parts0[w], parts1[w], loc);
	  break;
	case LT:
	case LTU:
	case GT:
	case GTU:
	  for (w = words - 1; w > 0; --w)
	    {
	      expand_cbranch (code, parts0[w], parts1[w], loc);
	      expand_cbranch (NE, parts0[w], parts1[w], fail);
	      code = unsigned_condition (code);
	    }
	  expand_cbranch (code, parts0[0], parts1[0], loc);
	  emit_label (fail);
	  break;
	case LE:
	case LEU:
	case GE:
	case GEU:
	  for (w = words - 1; w > 0; --w)
	    {
	      expand_cbranch (rcode, parts0[w], parts1[w], fail);
	      expand_cbranch (rcode, parts1[w], parts0[w], loc);
	      rcode = unsigned_condition (rcode);
	    }
	  expand_cbranch (unsigned_condition (code), parts0[0],
			  parts1[0], loc);
	  emit_label (fail);
	  break;

	default:
	  gcc_unreachable ();
	}
      return;
    }
  if (QImode == mode)
    emit_insn (gen_setccqi2 (op0, op1));
  else
    {
      gcc_assert (HImode == mode);
      emit_insn (gen_setcchi2 (op0, op1));
    }
  cmp = gen_rtx_fmt_ee (code, CCmode, cc0_rtx, const0_rtx);
  emit_jump_insn (gen_branchcc (loc, cmp));
}

void
msp430_expand_cbranch (rtx operands[])
{
  enum rtx_code code = GET_CODE (operands[0]);
  rtx op0 = operands[1];
  rtx op1 = operands[2];
  rtx loc = operands[3];

  gcc_assert (rtx_equal_p (XEXP (operands[0], 0), op0));
  gcc_assert (rtx_equal_p (XEXP (operands[0], 1), op1));
  expand_cbranch (code, op0, op1, loc);
}


const char*
msp430_output_branchcc (rtx insn, rtx operands[])
{
  rtx loc = operands[0];
  enum rtx_code cmp = GET_CODE (operands[1]);
  int dist = msp430_jump_dist (loc, insn);
  int in_range = MSP430_JUMP_IN_RANGE (dist);

  /* If optimization deleted a test instruction that ensured SR.V has
     the correct value for a jl/jge opcode, force the instruction to
     be emitted. */
  if ((cmp == LT || cmp == GE || cmp == LE || cmp == GT)
      && (!(cc_status.flags & (CC_EXPLICIT_COMPARE | CC_NO_OVERFLOW))))
    return 0;

  /* NB: Update the IF_THEN_ELSE block in adjust_pattern_length if
     this changes. */
  switch (cmp)
    {
    case EQ:
      return in_range ? "jeq\t%0" : "jne\t1f\n\tbr\t#%0\n1:\n";
    case NE:
      return in_range ? "jne\t%0" : "jeq\t1f\n\tbr\t#%0\n1:\n";
    case LT:
      return in_range ? "jl\t%0" : "jge\t1f\n\tbr\t#%0\n1:\n";
    case GE:
      return in_range ? "jge\t%0" : "jl\t1f\n\tbr\t#%0\n1:\n";
    case LTU:
      return in_range ? "jlo\t%0" : "jhs\t1f\n\tbr\t#%0\n1:\n";
    case GEU:
      return in_range ? "jhs\t%0" : "jlo\t1f\n\tbr\t#%0\n1:\n";
    case LE:
      return in_range ? "jeq\t%0\n\tjl\t%0" : "jeq\t1f\n\tjge\t2f\n1:\tbr\t#%0\n2:\n";
    case GT:
      return in_range ? "jeq\t1f\n\tjge\t%0\n1:\n" : "jeq\t2f\n\tjl\t2f\n1:\tbr\t#%0\n2:\n";
    case LEU:
      return in_range ? "jeq\t%0\n\tjlo\t%0" : "jeq\t1f\n\tjhs\t2f\n1:\tbr\t#%0\n2:\n";
    case GTU:
      return in_range ? "jeq\t1f\n\tjhs\t%0\n1:\n" : "jeq\t2f\n\tjlo\t2f\n1:\tbr\t#%0\n2:\n";
    default:
      break;
    }
  gcc_unreachable ();
  return 0;
}

static void
msp430_asm_function_epilogue (FILE *file,
			      HOST_WIDE_INT size ATTRIBUTE_UNUSED)
{
  struct machine_function *mfp = cfun->machine;
  if (mfp->inhibited_return_label != NULL)
    ASM_OUTPUT_LABEL(file, mfp->inhibited_return_label);
}

static void
msp430_asm_trampoline_template (FILE * fd)
{
  fprintf (fd, "; TRAMPOLINE HERE\n"
	   "; move context (either r1 or r4) to r6\n"
	   "; call function (0xf0f0 will be changed)\n");
  fprintf (fd, "\tmov	#0xf0f0, r6\n");
  fprintf (fd, "\tbr	#0xf0f0\n");
  fprintf (fd, "; END OF TRAMPOLINE\n\n");
}

static void
msp430_trampoline_init (rtx m_tramp, tree fndecl, rtx static_chain)
{
  rtx fnaddr = XEXP (DECL_RTL (fndecl), 0);
  emit_move_insn (gen_rtx_MEM (HImode, plus_constant (m_tramp, 2)),
		  static_chain);
  emit_move_insn (gen_rtx_MEM (HImode, plus_constant (m_tramp, 6)), fnaddr);
}

static void
msp430_init_libfuncs (void)
{
  libsym_mulqi3 = init_one_libfunc ("__mulqi3");
  set_optab_libfunc (smul_optab, QImode, "__mulqi3");

  libsym_mulqihi3 = init_one_libfunc ("__mulqihi3");
  set_optab_libfunc (smul_widen_optab, HImode, "__mulqihi3");
  libsym_umulqihi3 = init_one_libfunc ("__umulqihi3");
  set_optab_libfunc (umul_widen_optab, HImode, "__umulqihi3");
  libsym_mulhi3 = init_one_libfunc ("__mulhi3");
  set_optab_libfunc (smul_optab, HImode, "__mulhi3");

  libsym_mulhisi3 = init_one_libfunc ("__mulhisi3");
  set_optab_libfunc (smul_widen_optab, SImode, "__mulhisi3");
  libsym_umulhisi3 = init_one_libfunc ("__umulhisi3");
  set_optab_libfunc (umul_widen_optab, SImode, "__umulhisi3");
  libsym_mulsi3 = init_one_libfunc ("__mulsi3");
  set_optab_libfunc (smul_optab, SImode, "__mulsi3");

  libsym_mulsidi3 = init_one_libfunc ("__mulsidi3");
  set_optab_libfunc (smul_widen_optab, DImode, "__mulsidi3");
  libsym_umulsidi3 = init_one_libfunc ("__umulsidi3");
  set_optab_libfunc (umul_widen_optab, DImode, "__umulsidi3");
  libsym_muldi3 = init_one_libfunc ("__muldi3");
  set_optab_libfunc (smul_optab, DImode, "__muldi3");

  /* libcalls used for all divide and remainder operations */
  set_optab_libfunc (sdiv_optab, QImode, "__divqi3");
  set_optab_libfunc (sdiv_optab, HImode, "__divhi3");
  set_optab_libfunc (sdiv_optab, SImode, "__divsi3");
  set_optab_libfunc (sdiv_optab, DImode, "__divdi3");
  set_optab_libfunc (udiv_optab, QImode, "__udivqi3");
  set_optab_libfunc (udiv_optab, HImode, "__udivhi3");
  set_optab_libfunc (udiv_optab, SImode, "__udivsi3");
  set_optab_libfunc (udiv_optab, DImode, "__udivdi3");
  set_optab_libfunc (smod_optab, QImode, "__modqi3");
  set_optab_libfunc (smod_optab, HImode, "__modhi3");
  set_optab_libfunc (smod_optab, SImode, "__modsi3");
  set_optab_libfunc (smod_optab, DImode, "__moddi3");
  set_optab_libfunc (umod_optab, QImode, "__umodqi3");
  set_optab_libfunc (umod_optab, HImode, "__umodhi3");
  set_optab_libfunc (umod_optab, SImode, "__umodsi3");
  set_optab_libfunc (umod_optab, DImode, "__umoddi3");

  libsym_ashlqi3 = init_one_libfunc ("__ashlqi3");
  set_optab_libfunc (ashl_optab, QImode, "__ashlqi3");
  libsym_ashrqi3 = init_one_libfunc ("__ashrqi3");
  set_optab_libfunc (ashr_optab, QImode, "__ashrqi3");
  libsym_lshrqi3 = init_one_libfunc ("__lshrqi3");
  set_optab_libfunc (lshr_optab, QImode, "__lshrqi3");
  libsym_ashlhi3 = init_one_libfunc ("__ashlhi3");
  set_optab_libfunc (ashl_optab, HImode, "__ashlhi3");
  libsym_ashrhi3 = init_one_libfunc ("__ashrhi3");
  set_optab_libfunc (ashr_optab, HImode, "__ashrhi3");
  libsym_lshrhi3 = init_one_libfunc ("__lshrhi3");
  set_optab_libfunc (lshr_optab, HImode, "__lshrhi3");
  libsym_ashlsi3 = init_one_libfunc ("__ashlsi3");
  set_optab_libfunc (ashl_optab, SImode, "__ashlsi3");
  libsym_ashrsi3 = init_one_libfunc ("__ashrsi3");
  set_optab_libfunc (ashr_optab, SImode, "__ashrsi3");
  libsym_lshrsi3 = init_one_libfunc ("__lshrsi3");
  set_optab_libfunc (lshr_optab, SImode, "__lshrsi3");
  libsym_ashldi3 = init_one_libfunc ("__ashldi3");
  set_optab_libfunc (ashl_optab, DImode, "__ashldi3");
  libsym_ashrdi3 = init_one_libfunc ("__ashrdi3");
  set_optab_libfunc (ashr_optab, DImode, "__ashrdi3");
  libsym_lshrdi3 = init_one_libfunc ("__lshrdi3");
  set_optab_libfunc (lshr_optab, DImode, "__lshrdi3");

}

/* ---------------------------------------------------------------------------------------------------------------------------------------------- */

static bool
msp430_rtx_cost (rtx x, int code, int outer_code ATTRIBUTE_UNUSED,
		 int *total, bool speed)
{
  enum machine_mode mode = GET_MODE (x);
  int words = (GET_MODE_SIZE (mode) + UNITS_PER_WORD - 1) / UNITS_PER_WORD;
  bool is_430x = msp430_cpu & MSP430_CPU_MSP430X;
  bool rv;

  rv = true;
  switch (code)
    {
    case CONST_INT:
      {
	int cwords = words * UNITS_PER_WORD;
	cwords -= length_adjustment_for_int (INTVAL (x), mode);
	*total = cwords;
	break;
      }
    case SYMBOL_REF:
      *total = COSTS_N_INSNS (words);
      break;
    case ASHIFT:
    case ASHIFTRT:
    case LSHIFTRT:
      {
	rtx r_count = XEXP (x, 1);
	int count;

	if (CONST_INT_P (r_count))
	  count = INTVAL (r_count) & (GET_MODE_BITSIZE (mode) - 1);
	else
	  count = -1;

	if (msp430_enable_libcall_shift && speed && QImode != mode
	    && 0 > count)
	  break;

	if (0 > count)
	  {
	    *total =
	      COSTS_N_INSNS (shift_loop_overhead_words +
			     words * (1 << GET_MODE_SIZE (mode)));
	    break;
	  }

	*total = 0;
	while (0 < count)
	  {
	    if (GET_MODE_BITSIZE (mode) - 1 == count)
	      {
		*total += COSTS_N_INSNS (2);
		break;
	      }
	    if (count >= 48)
	      {
		*total += words * COSTS_N_INSNS (1);
		count -= 48;
		words = 1;
		continue;
	      }
	    if (count >= 32)
	      {
		*total += words * COSTS_N_INSNS (1);
		count -= 32;
		words = 2;
		continue;
	      }
	    if (count >= 16)
	      {
		*total += words * COSTS_N_INSNS (1);
		count -= 16;
	      }
	    if (count >= 8)
	      {
		*total +=
		  3 * (words - 1) * COSTS_N_INSNS (1) + COSTS_N_INSNS (2);
		count -= 8;
	      }
	    if (is_430x && HImode == mode
		&& MSP430_CPUX_MULTISHIFT_COUNT_P (count))
	      {
		while (0 < count)
		  {
		    *total += COSTS_N_INSNS (speed ? count - 1 : 1);
		    count -= MSP430_CPUX_MULTISHIFT_MAX;
		  }
		count = 0;
	      }
	    *total += count * words * COSTS_N_INSNS (1);
	    break;
	  }
	break;
      }

    case MULT:
      {
	rtx r_op2 = XEXP (x, 1);
	if (msp430_mpy & mpy_for_hardware_multiply (mode, GET_MODE (r_op2)))
	  break;
	*total = words * COSTS_N_INSNS (10);
	if (speed)
	  {
	    int mf = GET_MODE_BITSIZE (mode);

	    if (CONST_INT_P (r_op2))
	      {
		HOST_WIDE_INT ival =
		  trunc_int_for_mode (INTVAL (r_op2), mode);
		mf = 0;
		while (ival)
		  {
		    ++mf;
		    ival = (unsigned HOST_WIDE_INT) ival >> 2;
		  }
	      }
	    *total *= mf;
	  }
	break;
      }
    case DIV:
    case UDIV:
    case MOD:
    case UMOD:
      *total = words * COSTS_N_INSNS (20);
      if (speed)
	*total *= GET_MODE_BITSIZE (mode);
      break;
    default:
      *total *= words;
      rv = false;
      break;
    }

  return rv;
}

static int
msp430_address_cost (rtx x, bool speed)
{
  /* Nothing special now; at some point auto increment/decrement might
   * affect this. */
  return rtx_cost (x, MEM, speed);
}

/* Implement TARGET_OPTION_OPTIMIZATION_TABLE.  */
static const struct default_options msp430_option_optimization_table[] = {
  {OPT_LEVELS_1_PLUS, OPT_fomit_frame_pointer, NULL, 1},
  {OPT_LEVELS_NONE, 0, NULL, 0}
};

static void
msp430_globalize_label (FILE * stream, const char *name)
{
  if (*name == '*' || *name == '@')
    name++;
  if (*name >= '0' && *name <= '9')
    return;
  fputs (".global\t", stream);
  assemble_name (stream, name);
  putc ('\n', stream);
}

bool
msp430_inhibited_return_fallthru_p (rtx insn)
{
  struct machine_function *mfp;
  
  gcc_assert(cfun);
  mfp = cfun->machine;
  gcc_assert(mfp);
  gcc_assert(mfp->frame_flags & MSP430_FF_ready_for_return);
  gcc_assert(mfp->frame_flags & MSP430_FF_inhibit_return);

  insn = next_nonnote_insn (insn);
  gcc_assert (insn && BARRIER_P (insn));
  insn = next_nonnote_insn (insn);
  if (NULL_RTX == insn)
    return true;
  gcc_assert (LABEL_P (insn));
  return false;
}

static bool
msp430_cannot_modify_jumps_p (void)
{
  if (reload_completed && cfun && cfun->machine)
    return !!(cfun->machine->frame_flags & MSP430_FF_inhibit_return);
  return false;
}

/* 17.2 Driver -- no hooks */
/* 17.3 Run-time Target */
#undef TARGET_OPTION_OPTIMIZATION_TABLE
#define TARGET_OPTION_OPTIMIZATION_TABLE msp430_option_optimization_table

/* 17.4 Per-Function Data -- no hooks*/
/* 17.5: Storage Layout */

/* 17.6 Type Layout */
/* SF 3191528: TARGET_DEFAULT_SHORT_ENUMS */

/* 17.7 Registers */

/* 17.8 Register Classes */
/* TODO TARGET_IRA_COVER_CLASSES seems important */

/* 17.9 Old Constraints */

/* 17.10 Stack and Calling */
/* 17.10.5 Elimination */
#undef TARGET_FRAME_POINTER_REQUIRED
#define TARGET_FRAME_POINTER_REQUIRED msp430_frame_pointer_required
#undef TARGET_CAN_ELIMINATE
#define TARGET_CAN_ELIMINATE msp430_can_eliminate
/* 17.10.8 Scalar Return */
#undef TARGET_FUNCTION_VALUE
#define TARGET_FUNCTION_VALUE msp430_function_value
#undef TARGET_LIBCALL_VALUE
#define TARGET_LIBCALL_VALUE msp430_libcall_value
/* 17.19.9 Aggregate Return */
#undef TARGET_RETURN_IN_MEMORY
#define TARGET_RETURN_IN_MEMORY msp430_return_in_memory
/* 17.10.11 Function Entry */
#undef TARGET_ASM_FUNCTION_EPILOGUE
#define TARGET_ASM_FUNCTION_EPILOGUE msp430_asm_function_epilogue
/* 17.10.13 Tail Calls */

/* 17.11 Varargs */

/* 17.12 Trampolines */
#undef TARGET_ASM_TRAMPOLINE_TEMPLATE
#define TARGET_ASM_TRAMPOLINE_TEMPLATE msp430_asm_trampoline_template
#undef TARGET_TRAMPOLINE_INIT
#define TARGET_TRAMPOLINE_INIT msp430_trampoline_init

/* 17.13 Library Calls */
#undef TARGET_INIT_LIBFUNCS
#define TARGET_INIT_LIBFUNCS msp430_init_libfuncs

/* 17.14 Addressing Modes */
#undef TARGET_LEGITIMATE_ADDRESS_P
#define TARGET_LEGITIMATE_ADDRESS_P msp430_legitimate_address_p

/* 17.15 Anchored Addresses */

/* 17.16 Condition Code */

/* 17.17 Costs */
#undef TARGET_RTX_COSTS
#define TARGET_RTX_COSTS msp430_rtx_cost
#undef TARGET_ADDRESS_COST
#define TARGET_ADDRESS_COST msp430_address_cost

/* 17.18 Scheduling */

/* 17.19 Sections */
#undef TARGET_ASM_SELECT_SECTION
#define TARGET_ASM_SELECT_SECTION msp430_select_section

/* 17.20 PIC -- not supported */

/* 17.21 Assembler Format */
/* 17.21.1 File Framework */
#undef TARGET_ASM_FILE_START
#define TARGET_ASM_FILE_START msp430_file_start
#undef TARGET_ASM_FILE_START_FILE_DIRECTIVE
#define TARGET_ASM_FILE_START_FILE_DIRECTIVE true
/* TARGET_ASM_NAMED_SECTION is defined in msp430.h */
#undef 	TARGET_SECTION_TYPE_FLAGS
#define TARGET_SECTION_TYPE_FLAGS msp430_section_type_flags
/* 17.21.2 Data Output */
/* TODO: Verify these */
#undef TARGET_ASM_ALIGNED_HI_OP
#define TARGET_ASM_ALIGNED_HI_OP "\t.word\t"
/* 17.21.4 Label Output */
#undef TARGET_ASM_GLOBALIZE_LABEL
#define TARGET_ASM_GLOBALIZE_LABEL msp430_globalize_label
/* 17.21.9 Assembler Commands for Exception Regions */
#undef TARGET_EXCEPT_UNWIND_INFO
#define TARGET_EXCEPT_UNWIND_INFO sjlj_except_unwind_info

/* 17.25 Target Attributes */
#undef 	TARGET_ATTRIBUTE_TABLE
#define TARGET_ATTRIBUTE_TABLE msp430_attribute_table
#undef TARGET_OPTION_OVERRIDE
#define TARGET_OPTION_OVERRIDE msp430_option_override

/* 17.26 Emulated TLS */
/* 17.27 MIPS Coprocessors -- not relevant */
/* 17.28 PCH Target */
/* 17.29 C++ ABI */
/* 17.30 Named Address Spaces */

/* 17.31 Misc */
#undef TARGET_INIT_BUILTINS
#define TARGET_INIT_BUILTINS msp430_init_builtins
#undef TARGET_EXPAND_BUILTIN
#define TARGET_EXPAND_BUILTIN msp430_expand_builtin
#undef TARGET_SET_CURRENT_FUNCTION
#define TARGET_SET_CURRENT_FUNCTION msp430_set_current_function
#undef TARGET_CANNOT_MODIFY_JUMPS_P
#define TARGET_CANNOT_MODIFY_JUMPS_P msp430_cannot_modify_jumps_p

/* 17.1 Target Structure */
struct gcc_target targetm = TARGET_INITIALIZER;

#include "gt-msp430.h"
