;; -*- Mode: Scheme -*-
;;   Machine description for GNU compiler,
;;   for Texas Instruments msp430 MCUs
;;   Copyright (C) 2001, 2002, 2003 Free Software Foundation, Inc.
;;   Contributed by Dmitry Diky <diwil@mail.ru>
;;	 GCC 4.x port by Ivan Shcherbakov <mspgcc@sysprogs.org>

;; This work is partially financed by the European Commission under the
;; Framework 6 Information Society Technologies Project
;; "Wirelessly Accessible Sensor Populations (WASP)".

; This file is part of GCC.

;; GCC is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GCC is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GCC; see the file COPYING.  If not, write to
;; the Free Software Foundation, 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

(include "constraints.md")
(include "predicates.md")

;; Special characters after '%':
;;  A  No effect (add 0).
;;  B  Add 1 to REG number, 2 to MEM address or CONST_INT.
;;  C	   2		    4
;;  D	   3		    6
;;  S  Add 2 to memory address if using stack pointer (for pre_decr)

(define_constants
  [(REGNO_PC			0)
   (REGNO_SP			1)
   (REGNO_SR			2)
   (REGNO_FP			4)
   ]
  )

(define_constants
  [(UNSPEC_UNDEFINED		0)
   (UNSPEC_BITTEST_FOR_CARRY	6)
   (UNSPEC_PUSH_MULTI		20)
   (UNSPEC_POP_MULTI		21)
   (UNSPEC_LOAD_SP		105)
   ]
)

(define_constants
  [(UNSPECV_UNDEFINED			0)
   (UNSPECV_MPY_INHIBIT_INTR		100)
   (UNSPECV_MPY_RESTORE_INTR		101)
   (UNSPECV_NOP				200)
   (UNSPECV_DINT			201)
   (UNSPECV_EINT			202)
   (UNSPECV_READ_STATUS_REGISTER	203)
   (UNSPECV_WRITE_STATUS_REGISTER	204)
   (UNSPECV_BIC_STATUS_REGISTER		205)
   (UNSPECV_BIS_STATUS_REGISTER		206)
   (UNSPECV_BIC_SR_IRQ			207)
   (UNSPECV_BIS_SR_IRQ			208)
   (UNSPECV_READ_STACK_POINTER		209)
   (UNSPECV_WRITE_STACK_POINTER		210)
   (UNSPECV_DELAY_CYCLES		211)
   (UNSPECV_GET_INTERRUPT_STATE		212)
   (UNSPECV_SET_INTERRUPT_STATE		213)
   (UNSPECV_GET_WATCHDOG_CLEAR_VALUE	214)
   (UNSPECV_SET_WATCHDOG_CLEAR_VALUE	215)
   (UNSPECV_WATCHDOG_CLEAR		216)
   (UNSPECV_RLA				300)
   (UNSPECV_RRA				301)
   (UNSPECV_RLC				302)
   (UNSPECV_RRC				303)
   (UNSPECV_RRU				304)
   (UNSPECV_CLRC			305)
   (UNSPECV_MWSHL8XOR			310)
   (UNSPECV_MWSHR8XOR			311)
   ]
  )

; Instruction impact on condition code flags specific to MSP430
; comparison-based jump instructions.  Comparisons are equality (eq,
; ne), signed ordered (*gt, ge, lt, *le), and unsigned ordered (*gtu,
; geu, ltu, *leu) [* indicates a comparison not directly supported by
; a single msp430 instruction].  Where signed ordered are supported,
; so are unsigned ordered.
;
; undef : insn effect has not be categorized
; none : No change to flags
; clobber : Flags do not represent anything useful
; explicit : VNZC calculated explicitly by a comparison operation, as
;            opposed to as a side effect of an operation.  Side effect
;            calculations can be wrong for MSP430 conditional jumps
;            that reference the overflow flag (jge/jl).
; VNZC : All flags are valid for comparison between two values
; NZC : Flags are valid for equality and signed ordered comparisons
;       with zero.  Often C=~Z.  As used here, V is expected to be
;       cleared (CC_NO_OVERFLOW is set); thus this condition should
;       not be used for dadd and rlam which leave its value undefined.
; NZ : Flags are valid for signed ordered comparisons with zero.  V
;      and C are clobbered.  Used primarily in multi-word operations.
; N : Flags valid for signed ordered comparisons with zero, but not
;     for equality with zero.  V, C, and Z are clobbered.  Used
;     primarily in multi-word operations.
(define_attr "cc" "undef,none,clobber,explicit,VNZC,NZC,NZ,N"
  (const_string "undef"))

(define_attr "type" "branch,branch1,arith"
  (const_string "arith"))

(define_attr "msp430_has_hwmul" "yes,no"
  (const (if_then_else (symbol_ref "MSP430_HAS_HWMUL_INTERNAL")
		       (const_string "yes")
		       (const_string "no"))))

(define_attr "msp430_noint_hwmul" "" (symbol_ref "MSP430_NOINT_HWMUL"))

; Instruction formats.  Yes, it is irritating that format I (fmt1) has
; two operands, and format II (fmt2) has one.
;
; fmt1 : MSP430, three words : double-operand
; fmt2 : MSP430, two words : single-operand as dst
; fmt2s : MSP430, two words : single-operand as src
; emu1dd : MSP430 : fmt1 emulated dst, dst
; condjmp : MSP430, one word : conditional jump with 10-bit word offset
; fmt1x : MSP430X, four words : extension word, double-operand (fmt1)
; fmt2x : MSP430X, three words : extension word, single-operand (fmt2)
; fmtx : MSP430X, two words : no extension word, single-operand
(define_attr "instr_format"
  "undef,fmt1,fmt2,fmt2s,emu1dd,condjmp"
  (const_string "undef"))

; Multiplier for basic instr_format-based length when template includes
; multiple instances of the same instruction.
(define_attr "instr_mult"
  ""
  (cond [(match_operand:SI 0 "" "")
	 (const_int 2)
	 (match_operand:SF 0 "" "")
	 (const_int 2)
	 (match_operand:DI 0 "" "")
	 (const_int 4)]
	(const_int 1)))

; Length is calculated in bytes, and depends in most cases solely on
; the instruction format and the number of times the base instruction
; appears in the output template (for multi-word operands).
(define_attr "length" ""
  (cond [(eq_attr "instr_format" "fmt1,emu1dd")		; insn, src word, dst word
	 (mult (const_int 6) (attr "instr_mult"))
	 (eq_attr "instr_format" "fmt2,fmt2s")		; insn, dst word
	 (mult (const_int 4) (attr "instr_mult"))
	 (eq_attr "instr_format" "condjmp")		; insn (assume local jump)
	 (mult (const_int 2) (attr "instr_mult"))]
	(const_int 0)))

;; SI and SF as blobs are treated the same
(define_mode_iterator SISF [SI SF])

;; Modes to which QI can extend
(define_mode_iterator QIEXT [DI SI HI])

;; Modes to which HI can extend
(define_mode_iterator HIEXT [DI SI])

;; Integer modes that can be widened
(define_mode_iterator WidenableINTModes [SI HI QI])

;; Integer modes
(define_mode_iterator INTModes [DI SI HI QI])

;; Operand constraints for INTModes iterators
(define_mode_attr INTModes_constraint_dst
  [(QI "=rm")		       ; QImode goes to register or memory
   (HI "=rm")		       ; HImode goes to register or memory
   (SI "=rm")		       ; SImode goes to register or memory
   (DI "=rm")])		       ; DImode goes to register or memory

(define_mode_attr INTModes_constraint_src
  [(QI "UmQi")		       ; QI, HI allow anything
   (HI "UmQi")
   (SI "rmQi")		       ; SI, DI remove frame and stack pointer
   (DI "rmQi")])

(define_mode_attr INTModes_constraint_matchdst
  [(QI "0")
   (HI "0")
   (SI "0")
   (DI "0")])

; Variant for commutative operations
(define_mode_attr INTModes_commutative_matchdst
  [(QI "%0")
   (HI "%0")
   (SI "%0")
   (DI "%0")])

; Variants that allow operation on the frame and stack pointer for HImode
(define_mode_attr INTModes_frameok_dst
  [(QI "=rm")		       ; QImode goes to register or memory
   (HI "=Um")		       ; HImode allows frame and stack pointer
   (SI "=rm")		       ; SImode goes to register or memory
   (DI "=rm")])		       ; DImode goes to register or memory

(define_mode_attr INTModes_frameok_src
  [(QI "UmQi")		       ; QI, HI allow anything
   (HI "UmQi")		       ; SI, DI remove frame and stack pointer
   (SI "rmQi")
   (DI "rmQi")])

; For LSW-first arithmetic on values that might be multiple words
(define_mode_attr INTModes_cc_VNZC
  [(QI "VNZC")
   (HI "VNZC")
   (SI "N")
   (DI "N")])

(define_mode_attr INTModes_cc_NZC
  [(QI "NZC")
   (HI "NZC")
   (SI "N")
   (DI "N")])

(define_mode_attr INTModes_cc_NZ
  [(QI "NZ")
   (HI "NZ")
   (SI "N")
   (DI "N")])

; For MSW-first shift through carry on values that might be multiple words
(define_mode_attr INTModes_reversed_cc_NZ
  [(QI "NZ")
   (HI "NZ")
   (SI "clobber")
   (DI "clobber")])

; Next wider mode, in lower case
(define_mode_attr wider_mode
  [(QI "hi")
   (HI "si")
   (SI "di")])

; Next wider mode, in upper case
(define_mode_attr WIDER_MODE
  [(QI "HI")
   (HI "SI")
   (SI "DI")])

;; Integer modes that fit in a single register
(define_mode_iterator INTRegModes [HI QI])

;; ==================
;; Built-in functions

; nop()
(define_insn "nop"
  [(unspec_volatile [(const_int 0)] UNSPECV_NOP)]
  ""
  "nop"
  [(set_attr "cc" "none")
   (set_attr "length" "2")
   ])

; dint()
(define_insn "dint"
  [(unspec_volatile [(const_int 0)] UNSPECV_DINT)
   (clobber (reg:HI REGNO_SR))]
  ""
  "dint"
  [(set_attr "cc" "none")
   (set_attr "length" "2")
   ])

; eint()
(define_insn "eint"
  [(unspec_volatile [(const_int 0)] UNSPECV_EINT)
   (clobber (reg:HI REGNO_SR))]
  ""
  "eint"
  [(set_attr "cc" "none")
   (set_attr "length" "2")
   ])

; read_status_register(retloc)
(define_insn "read_status_register"
  [(set (match_operand:HI 0 "nonimmediate_operand" "=rm")
	(unspec_volatile:HI [(reg:HI REGNO_SR)] UNSPECV_READ_STATUS_REGISTER))]
  ""
  "mov\tr2, %0"
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt1")])

; get_interrupt_state(retloc)
; Yes, this is equivalent to read_status_register.
(define_insn "get_interrupt_state"
  [(set (match_operand:HI 0 "nonimmediate_operand" "=rm")
	(unspec_volatile:HI [(reg:HI REGNO_SR)] UNSPECV_GET_INTERRUPT_STATE))]
  ""
  "mov\tr2, %0"
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt1")])

; write_status_register(sr_value)
(define_insn "write_status_register"
  [(set (reg:HI REGNO_SR)
	(unspec_volatile:HI [(match_operand:HI 0 "msp430_source_operand" "rmQi")] UNSPECV_WRITE_STATUS_REGISTER))]
   ""
   "mov\t%0, r2"
   [(set_attr "cc" "clobber")
    (set_attr "instr_format" "fmt1")])

; set_interrupt_state(istate)
; Yes, this is equivalent to write_status_register.
(define_insn "set_interrupt_state"
  [(set (reg:HI REGNO_SR)
	(unspec_volatile:HI [(match_operand:HI 0 "msp430_source_operand" "rmQi")] UNSPECV_SET_INTERRUPT_STATE))]
   ""
   "mov\t%0, r2"
   [(set_attr "cc" "clobber")
    (set_attr "instr_format" "fmt1")])

; bic_status_register(sr_value)
(define_insn "bic_status_register"
  [(set (reg:HI REGNO_SR)
	(unspec_volatile:HI [(match_operand:HI 0 "msp430_source_operand" "rmQi")] UNSPECV_BIC_STATUS_REGISTER))]
   ""
   "bic\t%0, r2"
   [(set_attr "cc" "clobber")
    (set_attr "instr_format" "fmt1")])

; bis_status_register(sr_value)
(define_insn "bis_status_register"
  [(set (reg:HI REGNO_SR)
	(unspec_volatile:HI [(match_operand:HI 0 "msp430_source_operand" "rmQi")] UNSPECV_BIS_STATUS_REGISTER))]
   ""
   "bis\t%0, r2"
   [(set_attr "cc" "clobber")
    (set_attr "instr_format" "fmt1")])

; bic_status_register_on_exit(stack_reference, status_value)
(define_insn "bic_status_register_on_exit"
  [(set (match_operand:HI 0 "memory_operand" "=m")
	(unspec_volatile:HI [(match_operand:HI 1 "msp430_source_operand" "rmQi")] UNSPECV_BIC_SR_IRQ))]
   ""
   "bic\t%1, %0"
   [(set_attr "cc" "none")
    (set_attr "instr_format" "fmt1")])
 
; bis_status_register_on_exit(stack_reference, status_value)
(define_insn "bis_status_register_on_exit"
  [(set (match_operand:HI 0 "memory_operand" "=m")
	(unspec_volatile:HI [(match_operand:HI 1 "msp430_source_operand" "rmQi")] UNSPECV_BIS_SR_IRQ))]
   ""
   "bis\t%1, %0"
   [(set_attr "cc" "none")
    (set_attr "instr_format" "fmt1")])
 
; read_stack_pointer(retloc)
(define_insn "read_stack_pointer"
  [(set (match_operand:HI 0 "nonimmediate_operand" "=rm")
	(unspec_volatile:HI [(reg:HI REGNO_SP)] UNSPECV_READ_STACK_POINTER))]
   ""
   "mov\tr1, %0"
   [(set_attr "cc" "none")
    (set_attr "instr_format" "fmt1")])

; write_stack_pointer(sr_value)
(define_insn "write_stack_pointer"
  [(set (reg:HI REGNO_SP)
	(unspec_volatile:HI [(match_operand:HI 0 "msp430_source_operand" "rmQi")] UNSPECV_WRITE_STACK_POINTER))]
   ""
   "mov\t%0, r1"
   [(set_attr "cc" "none")
    (set_attr "instr_format" "fmt1")])

; delay_cycles_init(loopreg, count)
(define_insn "delay_cycles_init"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(unspec_volatile:HI [(match_operand:HI 1 "const_int_operand" "i")] UNSPECV_DELAY_CYCLES))
   (clobber (match_dup 0))]
  ""
  "mov\t%1, %0"
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt1")])

; delay_cycles_decr(loopreg)
(define_insn "delay_cycles_decr"
  [(set (match_operand:HI 0 "register_operand" "+r")
	(unspec_volatile:HI [(minus:HI (match_dup 0) (const_int 1))] UNSPECV_DELAY_CYCLES))]
  ""
  "dec\t%0"
  [(set_attr "cc" "VNZC")
   (set_attr "instr_format" "fmt1")])

; get_watchdog_clear_value(retloc)
(define_insn "get_watchdog_clear_value"
  [(set (match_operand:HI 0 "nonimmediate_operand" "=rm")
	(unspec_volatile:HI [(const_int 0)] UNSPECV_GET_WATCHDOG_CLEAR_VALUE))]
  ""
  "mov\t&__wdt_clear_value, %0"
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt1")])

; set_watchdog_clear_value(retloc, value)
(define_insn "set_watchdog_clear_value"
  [(unspec_volatile:HI [(match_operand:HI 0 "msp430_source_operand" "rmQi")] UNSPECV_SET_WATCHDOG_CLEAR_VALUE)]
  ""
  "mov\t%0, &__wdt_clear_value"
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt1")])
  
; watchdog_clear()
(define_insn "watchdog_clear"
  [(unspec_volatile [(const_int 0)] UNSPECV_WATCHDOG_CLEAR)]
  ""
  "mov\t&__wdt_clear_value, &__WDTCTL"
  [(set_attr "cc" "none")
   (set_attr "length" "6")
   ])

;; extendmn2 (sign extension)

(define_insn "extend8bithi1"
  [(set (match_operand:HI 0 "nonimmediate_operand" "+rmQ")
	(sign_extend:HI (subreg:QI (match_dup 0) 0)))]
  ""
  "sxt\t%0"
  [(set_attr "cc" "NZC")
   (set_attr "instr_format" "fmt2")])

(define_expand "extendsidi2"
  [(set (match_operand:DI 0 "nonimmediate_operand" "")
        (sign_extend:DI (match_operand:SI 1 "general_operand" "")))]
  ""
  "msp430_expand_signextend (operands); DONE;")
(define_expand "extendhi<mode>2"
  [(set (match_operand:HIEXT 0 "nonimmediate_operand" "")
        (sign_extend:<MODE> (match_operand:HI 1 "general_operand" "")))]
  ""
  "msp430_expand_signextend (operands); DONE;")
(define_expand "extendqi<mode>2"
  [(set (match_operand:QIEXT 0 "nonimmediate_operand" "")
        (sign_extend:<MODE> (match_operand:QI 1 "general_operand" "")))]
  ""
  "msp430_expand_signextend (operands); DONE;")

;; zero_extendmn2 (zero extension)

(define_expand "zero_extendsidi2"
  [(set (match_operand:DI 0 "nonimmediate_operand" "")
        (zero_extend:DI (match_operand:SI 1 "general_operand" "")))]
  ""
  "msp430_expand_zeroextend (operands); DONE;")
(define_expand "zero_extendhi<mode>2"
  [(set (match_operand:HIEXT 0 "nonimmediate_operand" "")
        (zero_extend:<MODE> (match_operand:HI 1 "general_operand" "")))]
  ""
  "msp430_expand_zeroextend (operands); DONE;")
(define_expand "zero_extendqi<mode>2"
  [(set (match_operand:QIEXT 0 "nonimmediate_operand" "")
        (zero_extend:<MODE> (match_operand:QI 1 "general_operand" "")))]
  ""
  "msp430_expand_zeroextend (operands); DONE;")

;; bswap (standard rtl, not standard insn)

(define_insn "*bswaphi2"
  [(set (match_operand:HI 0 "nonimmediate_operand" "=rm")
        (bswap:HI (match_operand:HI 1 "nonimmediate_operand" "0")))]
  ""
"swpb\t%0"
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt2")])

(define_expand "bswaphi1"
  [(set (match_operand:HI 0 "nonimmediate_operand" "+rm")
	(bswap:HI (match_dup 0)))]
  ""
  "")

;; movm

(define_insn "pushqi1"
  [(set (mem:QI (pre_modify:HI (reg:HI REGNO_SP) (plus:HI (reg:HI REGNO_SP) (const_int -2))))
	(match_operand:QI 0 "general_operand" "rmi"))]
  ""
  "push.b\t%S0"
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt2s")])

(define_insn "loadqi2"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(zero_extend:HI (match_operand:QI 1 "msp430_source_operand" "UmQi")))]
  ""
  "mov.b\t%1, %0"
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt1")])

(define_insn "storeqi2"
  [(set (match_operand:QI 0 "nonimmediate_operand" "=rm")
	(subreg:QI (match_operand:HI 1 "register_operand" "r") 0))]
  ""
  "mov.b\t%1, %0"
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt1")])

(define_insn "*movqi2"
  [(set (match_operand:QI 0 "nonimmediate_operand" "=rm")
	(match_operand:QI 1 "msp430_source_operand" "UmQi"))]
  ""
  "mov.b\t%1, %0"
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt1")])

(define_expand "movqi" 
  [(set (match_operand:QI 0 "nonimmediate_operand" "")
	(match_operand:QI 1 "general_operand" ""))]
  ""
  "")

; pushmhi2(num_regs, high_regno)
(define_insn "pushmhi2"
  [(set (reg:HI REGNO_SP)
	(unspec:HI [(match_operand:HI 0 "const_int_operand" "i")
		    (match_operand:HI 1 "const_int_operand" "N")]
		   UNSPEC_PUSH_MULTI))]
  ""
  {
    rtx xoperands[2];
    xoperands[0] = operands[0];
    xoperands[1] = gen_rtx_REG (HImode, INTVAL (operands[1]));
    output_asm_insn ("pushm\t%0, %1", xoperands);
    return "";
  }
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt2s")])

; popmhi2(num_regs, high_regno)
(define_insn "popmhi2"
  [(set (reg:HI REGNO_SP)
	(unspec:HI [(match_operand:HI 0 "const_int_operand" "i")
		    (match_operand:HI 1 "const_int_operand" "N")]
		   UNSPEC_POP_MULTI))]
  ""
  {
    rtx xoperands[2];
    xoperands[0] = operands[0];
    xoperands[1] = gen_rtx_REG (HImode, INTVAL (operands[1]));
    output_asm_insn ("popm\t%0, %1", xoperands);
    return "";
  }
  [(set_attr "cc" "none")		; not true if REGNO_SR covered
   (set_attr "instr_format" "fmt2s")])

(define_insn "pushhi1"
  [(set (mem:HI (pre_dec:HI (reg:HI REGNO_SP)))
	(match_operand:HI 0 "general_operand" "Wmi"))]
  ""
  "push\t%S0"
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt2s")])

(define_insn "pophi1"
  [(set (match_operand:HI 0 "nonimmediate_operand" "=r,w")
	(mem:HI (post_inc:HI (reg:HI REGNO_SP))))]
  ""
  "pop\t%0"
  [(set_attr "cc" "none,clobber")
   (set_attr "instr_format" "fmt1")])

(define_insn "*movhi2"
  [(set (match_operand:HI 0 "nonimmediate_operand" "=Um")
	(match_operand:HI 1 "msp430_source_operand" "UmQi"))]
  ""
  "mov\t%1, %0"
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt1")])

(define_expand "movhi" 
  [(set (match_operand:HI 0 "nonimmediate_operand" "")
        (match_operand:HI 1 "general_operand" ""))]
  ""
  "")

(define_insn "*pushm<mode>1"
  [(set (mem:SISF (pre_dec:HI (reg:HI REGNO_SP)))
	(match_operand:SISF 0 "register_operand" "r"))]
  "msp430_cpu & MSP430_CPU_MSP430X"
  "pushm\t#2, %B0"
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt2s")
   (set_attr "instr_mult" "1")])

(define_insn "push<mode>1"
  [(set (mem:SISF (pre_dec:HI (reg:HI REGNO_SP)))
	(match_operand:SISF 0 "general_operand" "ST,rmi"))]
  ""
  "@
push\t2+%S0\;push\t2+%S0
push\t%B0\;push\t%A0"
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt2s")
   (set_attr "instr_mult" "2")])

(define_insn "*mov<mode>2"
  [(set (match_operand:SISF 0 "nonimmediate_operand" "=rm")
	(match_operand:SISF 1 "msp430_source_operand" "rmQi"))]
  ""
  {
    return msp430_mov_noclobber (operands);
  }
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt1")])

(define_expand "mov<mode>" 
  [(set (match_operand:SISF 0 "nonimmediate_operand" "")
        (match_operand:SISF 1 "general_operand" ""))]
  ""
  "")

(define_insn "*pushmdi1"
  [(set (mem:DI (pre_dec:HI (reg:HI REGNO_SP)))
	(match_operand:DI 0 "register_operand" "r"))]
  "msp430_cpu & MSP430_CPU_MSP430X"
  "pushm\t#4, %D0"
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt2s")
   (set_attr "instr_mult" "1")])

(define_insn "pushdi1"
  [(set (mem:DI (pre_dec:HI (reg:HI REGNO_SP)))
	(match_operand:DI 0 "general_operand" "ST,rmi"))]
  ""
  "@
push\t6+%S0\;push\t6+%S0\;push\t6+%S0\;push\t6+%S0
push\t%D0\;push\t%C0\;push\t%B0\;push\t%A0"
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt2s")
   (set_attr "instr_mult" "4")])

(define_insn "*movdi2"
  [(set (match_operand:DI 0 "nonimmediate_operand" "=rm")
	(match_operand:DI 1 "msp430_source_operand" "rmQi"))]
  ""
  {
    return msp430_mov_noclobber (operands);
  }
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt1")])

(define_expand "movdi" 
  [(set (match_operand:DI 0 "nonimmediate_operand" "")
        (match_operand:DI 1 "general_operand" ""))]
  ""
  "")

;; addm3, subm3

(define_split
  [(set (match_operand:INTModes 0 "nonimmediate_operand" "")
	(plus:<MODE> (match_operand:<MODE> 1 "nonimmediate_operand" "")
		     (match_operand:<MODE> 2 "const_int_operand" "")))]
  "CONST_INT_P (operands[2]) && MSP430_CG_INT_P(-INTVAL (operands[2])) && !MSP430_CG_INT_P(INTVAL (operands[2]))"
  [(set (match_dup 0)
	(minus:<MODE> (match_dup 1)
		      (match_dup 2)))]
  { operands[2] = gen_int_mode (-INTVAL (operands[2]), <MODE>mode); })

(define_split
  [(set (match_operand:INTModes 0 "nonimmediate_operand" "")
	(minus:<MODE> (match_operand:<MODE> 1 "nonimmediate_operand" "")
		      (match_operand:<MODE> 2 "const_int_operand" "")))]
  "CONST_INT_P (operands[2]) && MSP430_CG_INT_P(-INTVAL (operands[2])) && !MSP430_CG_INT_P(INTVAL (operands[2]))"
  [(set (match_dup 0)
	(plus:<MODE> (match_dup 1)
		     (match_dup 2)))]
  { operands[2] = gen_int_mode (-INTVAL (operands[2]), <MODE>mode); })

(define_insn "*add<mode>3_std"
  [(set (match_operand:INTModes 0 "nonimmediate_operand" "<INTModes_frameok_dst>")
        (plus:<MODE> (match_operand:<MODE> 1 "nonimmediate_operand" "<INTModes_commutative_matchdst>")
		     (match_operand:<MODE> 2 "msp430_source_operand" "<INTModes_frameok_src>")))]
  ""
  { return msp430_output_template (<MODE>mode, 2, "add", "addc", NULL); }
  [(set_attr "cc" "<INTModes_cc_VNZC>")
   (set_attr "instr_format" "fmt1")])

(define_insn "*sub<mode>3_std"
  [(set (match_operand:INTModes 0 "nonimmediate_operand" "<INTModes_frameok_dst>")
        (minus:<MODE> (match_operand:<MODE> 1 "nonimmediate_operand" "<INTModes_constraint_matchdst>")
		      (match_operand:<MODE> 2 "msp430_source_operand" "<INTModes_frameok_src>")))]
  ""
  { return msp430_output_template (<MODE>mode, 2, "sub", "subc", NULL); }
  [(set_attr "cc" "<INTModes_cc_VNZC>")
   (set_attr "instr_format" "fmt1")])

(define_expand "add<mode>3" 
  [(set (match_operand:INTModes 0 "nonimmediate_operand" "")
        (plus:<MODE> (match_operand:<MODE> 1 "nonimmediate_operand" "")
		     (match_operand:<MODE> 2 "general_operand" "")))]
  ""
  "")

(define_expand "sub<mode>3" 
  [(set (match_operand:INTModes 0 "nonimmediate_operand" "")
        (minus:<MODE> (match_operand:<MODE> 1 "nonimmediate_operand" "")
		      (match_operand:<MODE> 2 "general_operand" "")))]
  ""
  "")

;; one_cmplm2 (invert)

(define_expand "one_cmpl<mode>2"
  [(set (match_operand:INTModes 0 "nonimmediate_operand" "")
	(not:<MODE> (match_operand:<MODE> 1 "nonimmediate_operand" "")))]
  ""
  {
    if (! rtx_equal_p (operands[0], operands[1]))
      {
	emit_move_insn (operands[0], operands[1]);
	operands[1] = operands[0];
      }
  })

(define_insn "*one_cmpl<mode>2_insn"
  [(set (match_operand:INTModes 0 "nonimmediate_operand" "=rm")
	(not:<MODE> (match_operand:<MODE> 1 "nonimmediate_operand" "0")))]
  ""
  { return msp430_output_template (<MODE>mode, 0, "inv", NULL, NULL); }
  [(set_attr "cc" "<INTModes_cc_VNZC>")
   (set_attr "instr_format" "fmt2")])

;; negm2

(define_insn "negsf2"
  [(set (match_operand:SF 0 "nonimmediate_operand" "=r,m")
	(neg:SF (match_operand:SF 1 "nonimmediate_operand" "0,0")))]
  ""
  "xor\t#0x8000, %B0"
  [(set_attr "cc" "clobber")
   (set_attr "length" "4,6")])

(define_expand "neg<mode>2"
  [(match_operand:INTModes 0 "nonimmediate_operand" "")
   (match_operand:<MODE> 1 "general_operand" "")]
  ""
  {
    emit_insn (gen_one_cmpl<mode>2 (operands[0], operands[1]));
    emit_insn (gen_add<mode>3 (operands[0], operands[0], const1_rtx));
    DONE;
  })

;; nandm3 (extension)

(define_insn "*nand<mode>3"
  [(set (match_operand:INTModes 0 "nonimmediate_operand" "<INTModes_constraint_dst>")
        (and:<MODE> (not:<MODE> (match_operand:<MODE> 1 "msp430_source_operand" "<INTModes_constraint_src>"))
		    (match_operand:<MODE> 2 "general_operand" "<INTModes_constraint_matchdst>")))]
  ""
  { return msp430_output_template (<MODE>mode, 1, "bic", NULL, NULL); }
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt1")])

; TODO peephole use bic with op2 inverted is CG-int
; TODO peephole andhi3 with 0xFF -> and.b or clr.b
; TODO convert multi-word to word, apply individual optimizations e.g. and -1

(define_insn "and<mode>3"
  [(set (match_operand:INTModes 0 "nonimmediate_operand" "<INTModes_constraint_dst>")
        (and:<MODE> (match_operand:<MODE> 1 "nonimmediate_operand" "<INTModes_commutative_matchdst>")
		    (match_operand:<MODE> 2 "msp430_source_operand" "<INTModes_constraint_src>")))]
  ""
  { return msp430_output_template (<MODE>mode, 2, "and", NULL, NULL); }
  [(set_attr "cc" "<INTModes_cc_NZC>")
   (set_attr "instr_format" "fmt1")])

(define_insn "ior<mode>3"
  [(set (match_operand:INTModes 0 "nonimmediate_operand" "<INTModes_constraint_dst>")
        (ior:<MODE> (match_operand:<MODE> 1 "nonimmediate_operand" "<INTModes_commutative_matchdst>")
		    (match_operand:<MODE> 2 "msp430_source_operand" "<INTModes_constraint_src>")))]
  ""
  { return msp430_output_template (<MODE>mode, 2, "bis", NULL, NULL); }
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt1")])

(define_insn "xor<mode>3"
  [(set (match_operand:INTModes 0 "nonimmediate_operand" "<INTModes_constraint_dst>")
        (xor:<MODE> (match_operand:<MODE> 1 "nonimmediate_operand" "<INTModes_commutative_matchdst>")
		    (match_operand:<MODE> 2 "msp430_source_operand" "<INTModes_constraint_src>")))]
  ""
  { return msp430_output_template (<MODE>mode, 2, "xor", NULL, NULL); }
  [(set_attr "cc" "<INTModes_cc_VNZC>")
   (set_attr "instr_format" "fmt1")])

;; absm2
(define_expand "abs<mode>2"
  [(match_operand:INTModes 0 "nonimmediate_operand" "<INTModes_constraint_dst>")
   (match_operand:<MODE> 1 "general_operand" "<INTModes_constraint_matchdst>")]
  ""
  {
    rtx test_ge = gen_rtx_GE (<MODE>mode, operands[0], const0_rtx);
    rtx dest_label = gen_label_rtx ();

    emit_move_insn (operands[0], operands[1]);
    emit_jump_insn (gen_cbranch<mode>4 (test_ge, operands[0], const0_rtx, dest_label));
    emit_insn (gen_neg<mode>2 (operands[0], operands[0]));
    emit_label (dest_label);
    DONE;
  })

(define_insn "abssf2"
  [(set (match_operand:SF 0 "nonimmediate_operand" "=r,m")
        (abs:SF (match_operand:SF 1 "nonimmediate_operand" "0,0")))]
  ""
  "and\t#0x7fff, %B0"
  [(set_attr "cc" "clobber,clobber")
   (set_attr "length" "2,3")])		; BAD LENGTH

;; strlenm

(define_expand "strlenhi"
  [(match_operand:HI 0 "nonimmediate_operand" "=rm")
   (match_operand:BLK 1 "memory_operand" "m")
   (match_operand:QI 2 "general_operand" "rmi")
   (match_operand:QI 3 "const_int_operand" "i")]
  ""
  {
    rtx sp_rtx = gen_reg_rtx (Pmode);
    rtx sp_deref_rtx = gen_rtx_MEM (QImode, sp_rtx);
    rtx guard_rtx = operands[2];
    rtx loop_label = gen_label_rtx ();
    rtx test_ne = gen_rtx_NE (QImode, sp_deref_rtx, guard_rtx);

    emit_move_insn (sp_rtx, XEXP (operands[1], 0));
    emit_insn (gen_subhi3 (sp_rtx, sp_rtx, const1_rtx));
    emit_label (loop_label);
    emit_insn (gen_addhi3 (sp_rtx, sp_rtx, const1_rtx));
    emit_jump_insn (gen_cbranchqi4 (test_ne, sp_deref_rtx, guard_rtx, loop_label));
    emit_insn (gen_subhi3 (operands[0], sp_rtx, XEXP (operands[1], 0)));
    DONE;
  })

;;========================================================================
;; compare

(define_expand "cbranch<mode>4"
  [(set (cc0) (compare
               (match_operand:INTModes 1 "nonimmediate_operand")
               (match_operand:<MODE> 2 "general_operand")))
   (set (pc)
        (if_then_else (match_operator 0 "comparison_operator"
                       [(cc0) (const_int 0)])
      (label_ref (match_operand 3 "" ""))
      (pc)))]
  ""
  { msp430_expand_cbranch (operands); DONE; })

(define_insn "setcc<mode>2"
  [(set (cc0) (compare
               (match_operand:INTRegModes 0 "nonimmediate_operand" "Um")
               (match_operand:<MODE> 1 "general_operand" "Umi")))]
  ""
  { return msp430_output_template (<MODE>mode, 1, "cmp", NULL, NULL); }
  [(set_attr "cc" "explicit")
   (set_attr "instr_format" "fmt1")])

(define_insn "*bittest<mode>2"
  [(set (cc0) (compare
               (and:INTRegModes
		(match_operand:<MODE> 0 "nonimmediate_operand" "Um")
		(match_operand:<MODE> 1 "general_operand" "Umi"))
	       (const_int 0)))]
  ""
  { return msp430_output_template (<MODE>mode, 1, "bit", NULL, NULL); }
  [(set_attr "cc" "<INTModes_cc_NZ>")
   (set_attr "instr_format" "fmt1")])

(define_insn "branchcc"
  [(set (pc)
	(if_then_else
	 (match_operator 1 "comparison_operator" [(cc0) (const_int 0)])
	 (label_ref (match_operand 0 "" ""))
	 (pc)))]
  ""
  { return msp430_output_branchcc (insn, operands); }
  [(set_attr "cc" "none")
   (set_attr "instr_format" "condjmp")])

;;============================================================================
;; call
;;

(define_expand "call"
  [(call (match_operand:HI 0 "general_operand" "")
         (match_operand:HI 1 "general_operand" ""))]
  ""
  "")

(define_insn "*call_insn"
  [(call (mem:HI (match_operand:HI 0 "general_operand" "rmi"))
         (match_operand:HI 1 "general_operand" "X"))]
  ""
  "call\t%S0"
  [(set_attr "cc" "clobber")
   (set_attr "instr_format" "fmt2s")])

(define_expand "call_value"
  [(set (match_operand 0 "register_operand" "")
        (call (match_operand:HI 1 "general_operand" "")
              (match_operand:HI 2 "general_operand" "")))]
  ""
  "")

(define_insn "*call_value_insn"
  [(set (match_operand 0 "register_operand" "=r")
	(call (mem:HI (match_operand:HI 1 "general_operand" "rmi"))
	      (match_operand:HI 2 "general_operand" "X")))]
  ""
  "call\t%S1"
  [(set_attr "cc" "clobber")
   (set_attr "instr_format" "fmt2s")])

;;========================================================================
;; Multiplication 

(define_insn "mpy_inhibit_intr"
  [(set (mem:HI (pre_dec:HI (reg:HI REGNO_SP)))
	(unspec_volatile:HI [(const_int 0)] UNSPECV_MPY_INHIBIT_INTR))]
  ""
  "push\tr2\;dint\;nop"
  [(set_attr "cc" "none")
   (set_attr "length" "6")])

(define_insn "mpy_restore_intr"
  [(set (mem:HI (post_inc:HI (reg:HI REGNO_SP)))
	(unspec_volatile:HI [(const_int 0)] UNSPECV_MPY_RESTORE_INTR))]
  ""
  "pop\tr2"
  [(set_attr "cc" "clobber")
   (set_attr "length" "2")])

(define_expand "mul<mode>3"
  [(match_operand:INTModes 0 "nonimmediate_operand" "")
   (match_operand:<MODE> 1 "nonimmediate_operand" "")
   (match_operand:<MODE> 2 "general_operand" "")]  
  ""
  "msp430_expand_mul (operands, 1); DONE;")

(define_expand "mul<mode><wider_mode>3"
  [(match_operand:<WIDER_MODE> 0 "nonimmediate_operand" "")
   (match_operand:WidenableINTModes 1 "nonimmediate_operand" "")
   (match_operand:<MODE> 2 "general_operand" "")]
  ""
  "msp430_expand_mul (operands, 1); DONE;")

(define_expand "umul<mode><wider_mode>3"
  [(match_operand:<WIDER_MODE> 0 "nonimmediate_operand" "")
   (match_operand:WidenableINTModes 1 "nonimmediate_operand" "")
   (match_operand:<MODE> 2 "general_operand" "")]
  ""
  "msp430_expand_mul (operands, 0); DONE;")

;; All division currently supported by optabs

;; Shift operations

(define_insn "trunchiqi"
  [(set (match_operand:QI 0 "nonimmediate_operand" "=r,m") 
	(truncate:QI (match_operand:HI 1 "register_operand" "r,r")))]
  ""
  "mov.b\t%1, %0"
  [(set_attr "cc" "none,none")
   (set_attr "length" "1,2")])		; BAD LENGTH

(define_insn "truncsihi"
  [(set (match_operand:HI 0 "register_operand" "=r") 
	(truncate:HI (match_operand:SI 1 "register_operand" "r")))]
  ""
  "mov\t%1, %0"
  [(set_attr "cc" "none")
   (set_attr "length" "1")])		; BAD LENGTH


(define_insn "truncsiqi"
  [(set (match_operand:QI 0 "register_operand" "=r") 
	(truncate:QI (match_operand:SI 1 "register_operand" "r")))]
  ""
  "mov.b\t%1, %0"
  [(set_attr "cc" "none")
   (set_attr "length" "1")])		; BAD LENGTH

(define_insn "truncdiqi"
  [(set (match_operand:QI 0 "register_operand" "=r") 
	(truncate:QI (match_operand:DI 1 "register_operand" "r")))]
  ""
  "mov.b\t%1, %0"
  [(set_attr "cc" "none")
   (set_attr "length" "1")])		; BAD LENGTH

(define_insn "truncdisi"
  [(set (match_operand:SI 0 "register_operand" "=r") 
	(truncate:SI (match_operand:DI 1 "register_operand" "r")))]
  ""
  "mov\t%A1, %A0\;mov\t%B1, %B0"
  [(set_attr "cc" "none")
   (set_attr "length" "2")])		; BAD LENGTH

(define_expand "ashl<mode>3"
  [(set (match_operand:INTModes 0 "nonimmediate_operand" "")
	(ashift:<MODE> (match_operand:<MODE> 1 "nonimmediate_operand" "")
		       (match_operand:QI 2 "general_operand" "")))]
  ""
  "msp430_expand_ashl (operands); DONE;")

(define_expand "ashr<mode>3"
  [(set (match_operand:INTModes 0 "nonimmediate_operand" "")
	(ashiftrt:<MODE> (match_operand:<MODE> 1 "nonimmediate_operand" "")
			 (match_operand:QI 2 "general_operand" "")))]
  ""
  "msp430_expand_ashr (operands); DONE;")

(define_expand "lshr<mode>3"
  [(set (match_operand:INTModes 0 "nonimmediate_operand" "")
	(lshiftrt:<MODE> (match_operand:<MODE> 1 "nonimmediate_operand" "")
			 (match_operand:QI 2 "general_operand" "")))]
  ""
  "msp430_expand_lshr (operands); DONE;")

;; Miscellaneous shift helpers
(define_insn "clrc"
  [(unspec_volatile [(const_int 0)] UNSPECV_CLRC)]
  ""
  "clrc"
  [(set_attr "cc" "clobber")
   (set_attr "length" "2")
   ])

(define_insn "mwshl8xor"
  [(set (match_operand:HI 0 "register_operand" "+r")
	(unspec_volatile:HI [(match_operand:HI 1 "register_operand" "r")] UNSPECV_MWSHL8XOR))
   (use (match_dup 0))]
  ""
  "xor.b\t%1, %0\;xor\t%1, %0\;swpb\t%0"
  [(set_attr "cc" "clobber")
   (set_attr "instr_format" "fmt2")])

(define_insn "mwshr8xor"
  [(set (match_operand:HI 0 "register_operand" "+r")
	(unspec_volatile:HI [(match_operand:HI 1 "register_operand" "+r")] UNSPECV_MWSHR8XOR))
   (use (match_dup 0))
   (use (match_dup 1))
   (clobber (match_dup 1))]
  ""
  "swpb\t%1\;xor.b\t%1, %0\;xor\t%1, %0"
  [(set_attr "cc" "NZ")
   (set_attr "instr_format" "fmt2")])

;; Arithmetic shift left multiple positions

(define_expand "ashlhi_m"
  [(set (match_operand:HI 0 "nonimmediate_operand" "")
	(ashift:HI (match_operand:HI 1 "nonimmediate_operand" "")
		     (match_operand:QI 2 "const_int_operand" "")))]
  ""
  "")

(define_insn "*ashlhi_m_insn"
  [(set (match_operand:HI 0 "nonimmediate_operand" "=r")
	(ashift:HI (match_operand:HI 1 "nonimmediate_operand" "0")
		   (match_operand:QI 2 "const_int_operand" "M")))]
  "msp430_cpu & MSP430_CPU_MSP430X && CONST_INT_P(operands[2]) && MSP430_CPUX_MULTISHIFT_COUNT_P(INTVAL (operands[2]))"
  "rlam\t%2, %0"
  [(set_attr "cc" "NZ")
   (set_attr "instr_format" "fmt2")])

;; Arithmetic shift right multiple positions

(define_expand "ashrhi_m"
  [(set (match_operand:HI 0 "nonimmediate_operand" "")
	(ashiftrt:HI (match_operand:HI 1 "nonimmediate_operand" "")
		     (match_operand:QI 2 "const_int_operand" "")))]
  ""
  "")

(define_insn "*ashrhi_m_insn"
  [(set (match_operand:HI 0 "nonimmediate_operand" "=r")
	(ashiftrt:HI (match_operand:HI 1 "nonimmediate_operand" "0")
		     (match_operand:QI 2 "const_int_operand" "M")))]
  "msp430_cpu & MSP430_CPU_MSP430X && CONST_INT_P(operands[2]) && MSP430_CPUX_MULTISHIFT_COUNT_P(INTVAL (operands[2]))"
  "rram\t%2, %0"
  [(set_attr "cc" "NZ")
   (set_attr "instr_format" "fmt2")])

(define_insn "ashrhi_15"
  [(set (match_operand:HI 0 "nonimmediate_operand" "=r")
	(ashiftrt:HI (match_operand:HI 1 "nonimmediate_operand" "0")
		     (const_int 15)))]
  ""
  "swpb\t%0\;sxt\t%0\;swpb\t%0\;sxt\t%0"
  [(set_attr "cc" "clobber")
   (set_attr "instr_mult" "4")
   (set_attr "instr_format" "fmt2")])

;; Logical shift right multiple positions (CPUX, HI only)

(define_expand "lshrhi_m"
  [(set (match_operand:HI 0 "nonimmediate_operand" "")
	(lshiftrt:HI (match_operand:HI 1 "nonimmediate_operand" "")
		     (match_operand:QI 2 "const_int_operand" "")))]
  ""
  "")

(define_insn "*lshrhi_1_x"
  [(set (match_operand:HI 0 "nonimmediate_operand" "+r")
	(lshiftrt:HI (match_dup 0) (const_int 1)))]
  "msp430_cpu & MSP430_CPU_MSP430X"
  "rrum\t#1, %0"
  [(set_attr "cc" "NZ")
   (set_attr "instr_format" "fmt2")])

(define_insn "*lshrhi_m_insn"
  [(set (match_operand:HI 0 "nonimmediate_operand" "=r")
	(lshiftrt:HI (match_operand:HI 1 "nonimmediate_operand" "0")
		     (match_operand:QI 2 "const_int_operand" "M")))]
  "msp430_cpu & MSP430_CPU_MSP430X && CONST_INT_P(operands[2]) && MSP430_CPUX_MULTISHIFT_COUNT_P(INTVAL (operands[2]))"
  "rrum\t%2, %0"
  [(set_attr "cc" "NZ")
   (set_attr "instr_format" "fmt2")])

;; Basic single-position shifts

(define_insn "ashl<mode>_1"
  [(set (match_operand:INTModes 0 "nonimmediate_operand" "+rm")
	(ashift:<MODE> (match_dup 0) (const_int 1)))
   (clobber (reg:CC REGNO_SR))]
  ""
  { return msp430_output_template (<MODE>mode, 0, "rla", "rlc", NULL); }
  [(set_attr "cc" "<INTModes_cc_NZ>")
   (set_attr "instr_format" "emu1dd")])        ; emulated add dst, dst
  
(define_insn "rla<mode>1"
  [(set (match_operand:INTRegModes 0 "nonimmediate_operand" "+rm")
	(ashift:<MODE>
	 (unspec_volatile [(match_dup 0)] UNSPECV_RLA)
	 (const_int 1)))]
  ""
  { return msp430_output_template (<MODE>mode, 0, "rla", NULL, NULL); }
  [(set_attr "cc" "<INTModes_cc_NZ>")
   (set_attr "instr_format" "emu1dd")])        ; emulated add dst, dst

(define_insn "rlc<mode>1"
  [(set (match_operand:INTRegModes 0 "nonimmediate_operand" "+rm")
	(unspec_volatile:<MODE> [(match_dup 0)] UNSPECV_RLC))
   (use (reg:CC REGNO_SR))
   (clobber (reg:CC REGNO_SR))]
  ""
  { return msp430_output_template (<MODE>mode, 0, "rlc", NULL, NULL); }
  [(set_attr "cc" "<INTModes_cc_NZ>")
   (set_attr "instr_format" "emu1dd")])        ; emulated add dst, dst

(define_insn "ashr<mode>_1"
  [(set (match_operand:INTModes 0 "nonimmediate_operand" "+rm")
	(ashiftrt:<MODE> (match_dup 0) (const_int 1)))
   (clobber (reg:CC REGNO_SR))]
  ""
  { return msp430_output_reverse_template (<MODE>mode, 0, "rra", "rrc", NULL); }
  [(set_attr "cc" "<INTModes_reversed_cc_NZ>")
   (set_attr "instr_format" "fmt2")])
  
(define_insn "rra<mode>1"
  [(set (match_operand:INTRegModes 0 "nonimmediate_operand" "+rm")
	(ashiftrt:<MODE>
	 (unspec_volatile [(match_dup 0)] UNSPECV_RRA)
	 (const_int 1)))]
  ""
  { return msp430_output_reverse_template (<MODE>mode, 0, "rra", NULL, NULL); }
  [(set_attr "cc" "<INTModes_reversed_cc_NZ>")
   (set_attr "instr_format" "fmt2")])
  
(define_insn "rrc<mode>1"
  [(set (match_operand:INTRegModes 0 "nonimmediate_operand" "+rm")
	(unspec_volatile:<MODE> [(match_dup 0)] UNSPECV_RRC))
   (use (reg:CC REGNO_SR))
   (clobber (reg:CC REGNO_SR))]
  ""
  { return msp430_output_reverse_template (<MODE>mode, 0, "rrc", NULL, NULL); }
  [(set_attr "cc" "<INTModes_reversed_cc_NZ>")
   (set_attr "instr_format" "fmt2")])
  
(define_insn "lshr<mode>_1"
  [(set (match_operand:INTModes 0 "nonimmediate_operand" "+rm")
	(lshiftrt:<MODE> (match_dup 0) (const_int 1)))
   (clobber (reg:CC REGNO_SR))]
  ""
  { return msp430_output_reverse_template (<MODE>mode, 0, "rrc", NULL, "clrc"); }
  [(set_attr "cc" "<INTModes_reversed_cc_NZ>")
   (set_attr "instr_format" "fmt2")])

(define_insn "*rruhi1_x"
  [(set (match_operand:HI 0 "nonimmediate_operand" "+r")
	(lshiftrt:HI
	 (unspec_volatile [(match_dup 0)] UNSPECV_RRU)
	 (const_int 1)))]
  "msp430_cpu & MSP430_CPU_MSP430X"
  "rrum\t#1, %0"
  [(set_attr "cc" "NZ")
   (set_attr "instr_format" "fmt2")])

(define_insn "rru<mode>1"
  [(set (match_operand:INTRegModes 0 "nonimmediate_operand" "+rm")
	(lshiftrt:<MODE>
	 (unspec_volatile [(match_dup 0)] UNSPECV_RRU)
	 (const_int 1)))
   (clobber (reg:CC REGNO_SR))]
  ""
  { return msp430_output_reverse_template (<MODE>mode, 0, "rrc", NULL, "clrc"); }
  [(set_attr "cc" "<INTModes_reversed_cc_NZ>")
   (set_attr "instr_format" "fmt2")])

;; =====================================================================
;; single bit extract

(define_expand "extzv"
  [(match_operand:HI 0 "nonimmediate_operand" "") ; destination
   (match_operand:QI 1 "nonimmediate_operand" "") ; packed value
   (match_operand:HI 2 "const_int_operand" "")	  ; width
   (match_operand:HI 3 "const_int_operand" "")]	  ; starting bit
  ""
  {
    if (msp430_expand_extract (operands, false))
      DONE;
    FAIL;
  })

(define_insn "bittestforcarry<mode>2"
  [(set
    (reg:CC REGNO_SR)
    (unspec:CC
     [(match_operand:INTRegModes 0 "nonimmediate_operand" "Um")
      (match_operand:<MODE> 1 "general_operand" "Umi")]
     UNSPEC_BITTEST_FOR_CARRY))]
  ""
  { return msp430_output_template (<MODE>mode, 1, "bit", NULL, NULL); }
  [(set_attr "cc" "<INTModes_cc_NZC>")
   (set_attr "instr_format" "fmt1")])

;;=======================================================================
;;  various BRANCH insns...

;; Unconditional jump instruction.
(define_insn "jump"
  [(set (pc) (label_ref (match_operand 0 "" "")))]
  ""
  {
    int dist = msp430_jump_dist(operands[0],insn);
    if (dist<500 && dist>-500)
      return "jmp\t%0";
    return "br\t#%0";
  }
  [(set_attr "cc" "none")
   (set_attr "length" "2")])

; indirect jump
(define_insn "indirect_jump"
  [(set (pc) (match_operand:HI 0 "general_operand" "rmi"))]
  ""
  "br\t%0"
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt1")])


;; Table helper
(define_insn "tablejump"
  [(set (pc) (match_operand:HI 0 "general_operand" "rRP,i,m"))
   (use (label_ref (match_operand 1 "" "")))]
  ""
  "br\t%0\t; %1"
  [(set_attr "cc" "none")
   (set_attr "instr_format" "fmt1")])

;; return insn
(define_insn "*return_inhibited_fallthru"
  [(return)]
  "(reload_completed && cfun && cfun->machine
    && (cfun->machine->frame_flags & MSP430_FF_ready_for_return)
    && (cfun->machine->frame_flags & MSP430_FF_inhibit_return)
    && msp430_inhibited_return_fallthru_p (insn))"
  ""
  [(set_attr "cc" "none")
   (set_attr "length" "0")])

(define_insn "*return_inhibited_no_fallthru"
  [(return)]
  "(reload_completed && cfun && cfun->machine
    && (cfun->machine->frame_flags & MSP430_FF_ready_for_return)
    && (cfun->machine->frame_flags & MSP430_FF_inhibit_return))"
  {
    struct machine_function *mfp;
    gcc_assert (cfun);
    mfp = cfun->machine;
    gcc_assert (mfp);
    gcc_assert (NULL != mfp->inhibited_return_label);
    fprintf (asm_out_file, "\tbr\t#");
    assemble_name (asm_out_file, mfp->inhibited_return_label);
    return "\n";
  }
  [(set_attr "cc" "none")
   (set_attr "length" "4")])

(define_insn "*return_pop_sr"
  [(return)]
  "(reload_completed && cfun && cfun->machine
    && (cfun->machine->frame_flags & MSP430_FF_ready_for_return)
    && (cfun->machine->frame_flags & MSP430_FF_use_reti))"
  "reti"
  [(set_attr "cc" "none")
   (set_attr "length" "2")])

(define_insn "return"
  [(return)]
  "(reload_completed && cfun && cfun->machine
    && (cfun->machine->frame_flags & MSP430_FF_ready_for_return))"
  "ret"
  [(set_attr "cc" "none")
   (set_attr "length" "2")])

(define_expand "prologue"
  [(const_int 0)]
  ""
  "msp430_expand_prologue (); DONE;")

(define_expand "epilogue"
  [(const_int 0)]
  ""
  "msp430_expand_epilogue (); DONE;")

(include "peephole.md")
