;; -*- Mode: Scheme -*-
;;   Machine description for GNU compiler,
;;   for Texas Instruments msp430 MCUs
;;   Copyright (C) 2001, 2002, 2003 Free Software Foundation, Inc.
;;   Contributed by Dmitry Diky <diwil@mail.ru>
;;	 GCC 4.x port by Ivan Shcherbakov <mspgcc@sysprogs.org>

;; This work is partially financed by the European Commission under the
;; Framework 6 Information Society Technologies Project
;; "Wirelessly Accessible Sensor Populations (WASP)".

; This file is part of GCC.

;; GCC is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GCC is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GCC; see the file COPYING.  If not, write to
;; the Free Software Foundation, 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;; Gratuitous route through register for RMW on memory with reg operand
;; mov &m1, r0; op g3, r0; mov r0, &m1 => op g3, &m1 [r0 dead]
; see testsuite volpeep_reg.c
(define_peephole2
  [(set (match_operand:INTRegModes 0 "register_operand" "") 
	(match_operand:<MODE> 1 "nonimmediate_operand" ""))
   (set (match_dup 0) 
	(match_operator:<MODE> 2 "msp430_rmw_operator"
			       [(match_dup 0)
				(match_operand:<MODE> 3 "general_operand" "")]))
   (set (match_dup 1) (match_dup 0))]
  "peep2_reg_dead_p (3, operands[0])"
  [(set (match_dup 1) 
	(match_op_dup 2 [(match_dup 1) (match_dup 3)]))]
  "")

;; Gratuitous route through register for RMW on memory with non-reg operand
;; mov &m1, r0; mov &m3, r2; op r0, r2; mov r2, &m1 => op &m3, &m1 [r0 dead, r2 dead]
; see testsuite volpeep_volmem.c
(define_peephole2
  [(set (match_operand:INTRegModes 0 "register_operand" "") 
	(match_operand:<MODE> 1 "nonimmediate_operand" ""))
   (set (match_operand:<MODE> 2 "register_operand" "") 
	(match_operand:<MODE> 3 "nonimmediate_operand" ""))
   (set (match_dup 2) 
	(match_operator:<MODE> 4 "msp430_rmw_operator"
			       [(match_dup 2)
				(match_dup 0)]))
   (set (match_dup 1) (match_dup 2))]
  "peep2_reg_dead_p (4, operands[0]) && peep2_reg_dead_p (4, operands[2])"
  [(set (match_dup 1) 
	(match_op_dup 4 [(match_dup 1) (match_dup 3)]))]
  "")

;; Same as above but with operands swapped (used for sub)
;; mov &m1, r0; mov &m3, r2; op r2, r0; mov r0, &m1 => op &m3, &m1
; see testsuite volpeep_volmem.c
(define_peephole2
  [(set (match_operand:INTRegModes 0 "register_operand" "") 
	(match_operand:<MODE> 1 "nonimmediate_operand" ""))
   (set (match_operand:<MODE> 2 "register_operand" "") 
	(match_operand:<MODE> 3 "nonimmediate_operand" ""))
   (set (match_dup 0) 
	(match_operator:<MODE> 4 "msp430_rmw_operator"
			       [(match_dup 0)
				(match_dup 2)]))
   (set (match_dup 1) (match_dup 0))]
  "peep2_reg_dead_p (4, operands[0]) && peep2_reg_dead_p (4, operands[2])"
  [(set (match_dup 1) 
	(match_op_dup 4 [(match_dup 1) (match_dup 3)]))]
  "")

;; Assignment to volatile memory through a cast constant pointer
;; mov #c1, r0; mov &m3, r2; op4 g5, r2; mov r2, @r0 => op g4, &c1 [r0 dead, r2 dead]
; see testsuite vwa4.c
(define_peephole2
  [(set (match_operand 0 "pmode_register_operand" "") 
	(match_operand 1 "immediate_operand" ""))
   (set (match_operand:INTRegModes 2 "register_operand" "") 
	(match_operand:<MODE> 3 "nonimmediate_operand" ""))
   (set (match_dup 2) 
	(match_operator:<MODE> 4 "msp430_rmw_operator"
			       [(match_dup 2)
				(match_operand:<MODE> 5 "general_operand" "")]))
   (set (mem:<MODE> (match_dup 0)) (match_dup 2))]
  "peep2_reg_dead_p (4, operands[0]) && peep2_reg_dead_p (4, operands[2])"
  [(set (mem:<MODE> (match_dup 1))
	(match_op_dup 4 [(mem:<MODE> (match_dup 1)) (match_dup 5)]))]
  "")


;; mov &m1, r0; mov &m3, r2; op r2, r0; mov r0, &m1 => mov &m3, r2; op r2, &m1 [r0 dead, r2 live]
; see testsuite volpeep_mem.c; vwa_error.c
(define_peephole2
  [(set (match_operand:INTRegModes 0 "register_operand" "") 
	(match_operand:<MODE> 1 "nonimmediate_operand" ""))
   (set (match_operand:<MODE> 2 "register_operand" "") 
	(match_operand:<MODE> 3 "nonimmediate_operand" ""))
   (set (match_dup 0)
	(match_operator:<MODE> 4 "msp430_rmw_operator"
			       [(match_dup 0)
				(match_dup 2)]))
   (set (match_dup 1) (match_dup 0))]
  "peep2_reg_dead_p (4, operands[0]) && ! MEM_VOLATILE_P (operands[3]) && ! peep2_reg_dead_p (4, operands[2])"
  [(set (match_dup 2) (match_dup 3))
   (set (match_dup 1) 
	(match_op_dup 4 [(match_dup 1) (match_dup 2)]))]
  "")

;; mov &m1, r0; mov &m3, r2; nand r2, r0; mov r0, &m1 => nand &m3, &m1 [r0 dead, r2 dead]
; see testsuite volpeep_volmem.c
(define_peephole2
  [(set (match_operand:INTRegModes 0 "register_operand" "") 
	(match_operand:<MODE> 1 "nonimmediate_operand" ""))
   (set (match_operand:<MODE> 2 "register_operand" "") 
	(match_operand:<MODE> 3 "nonimmediate_operand" ""))
   (set (match_dup 0) 
	(and:<MODE> (not:<MODE> (match_dup 2))
		    (match_dup 0)))
   (set (match_dup 1) (match_dup 0))]
  "peep2_reg_dead_p (4, operands[0]) && peep2_reg_dead_p (4, operands[2])"
  [(set (match_dup 1) 
	(and:<MODE> (not:<MODE> (match_dup 3)) (match_dup 1)))]
  "")

;; mov &m1, r0; nand g2, r0; mov r0, &m1 => nand g2, &m1 [r0 dead]
; see testsuite volpeep_reg.c
(define_peephole2
  [(set (match_operand:INTRegModes 0 "register_operand" "") 
	(match_operand:<MODE> 1 "nonimmediate_operand" ""))
   (set (match_dup 0) 
	(and:<MODE> (not:<MODE> (match_operand:<MODE> 2 "general_operand" ""))
		    (match_dup 0)))
   (set (match_dup 1) (match_dup 0))]
  "peep2_reg_dead_p (3, operands[0])"
  [(set (match_dup 1) 
	(and:<MODE> (not:<MODE> (match_dup 2)) (match_dup 1)))]
  "")

;; mov &m1, r0; op r0, x2 => op &m1, x2 [r0 dead]
; transiently checked with vwa5.c
(define_peephole2
  [(set (match_operand:INTRegModes 0 "register_operand" "") 
	(match_operand:<MODE> 1 "nonimmediate_operand" ""))
   (set (match_operand:<MODE> 2 "nonimmediate_operand" "")
	(match_operator:<MODE> 3 "msp430_rmw_operator"
			       [(match_dup 2)
				(match_dup 0)]))]
  "peep2_reg_dead_p (2, operands[0])"
  [(set (match_dup 2) 
	(match_op_dup 3 [(match_dup 2) (match_dup 1)]))]
  "")

;; mov &m1, r0; nand r0, &m2 => nand &m1, &m2 [r0 dead]
; see testsuite volpeep_volmem.c
(define_peephole2
  [(set (match_operand:INTRegModes 0 "register_operand" "") 
	(match_operand:<MODE> 1 "nonimmediate_operand" ""))
   (set (match_operand:<MODE> 2 "nonimmediate_operand" "")
	(and:<MODE> (not:<MODE> (match_dup 0))
		    (match_dup 2)))]
  "peep2_reg_dead_p (2, operands[0])"
  [(set (match_dup 2) 
	(and:<MODE> (not:<MODE> (match_dup 1)) (match_dup 2)))]
  "")

;; mov &m1, r0; mov r0 &m2; => mov &m1, &m2 [r0 dead]
; see testsuite volpeep_volmem.c
(define_peephole2
  [(set (match_operand:INTRegModes 0 "register_operand" "")
        (match_operand:<MODE> 1 "general_operand" ""))
   (set (match_operand:<MODE> 2 "nonimmediate_operand" "")
        (match_dup 0))]
  "peep2_reg_dead_p (2, operands[0])"
  [(set (match_dup 2) (match_dup 1))]
  "")

;; mov $r2, r0 ; mov r0, g2 => mov $r2, g2 [r0 dead]
; see testsuite builtins_read_sr.c
(define_peephole2
  [(set (match_operand:HI 0 "register_operand" "")
	(unspec_volatile:HI [(reg:HI REGNO_SR)] UNSPECV_READ_STATUS_REGISTER))
   (set (match_operand:HI 2 "nonimmediate_operand" "")
	(match_dup 0))]
  "peep2_reg_dead_p (2, operands[0])"
  [(set (match_dup 2) (unspec_volatile:HI [(reg:HI REGNO_SR)] UNSPECV_READ_STATUS_REGISTER))]
  "")

;; mov $r1, r0 ; mov r0, g2 => mov $r1, g2 [r0 dead]
; see testsuite builtins_sp.c
(define_peephole2
  [(set (match_operand:HI 0 "register_operand" "")
	(unspec_volatile:HI [(reg:HI REGNO_SP)] UNSPECV_READ_STACK_POINTER))
   (set (match_operand:HI 2 "nonimmediate_operand" "")
	(match_dup 0))]
  "peep2_reg_dead_p (2, operands[0])"
  [(set (match_dup 2) (unspec_volatile:HI [(reg:HI REGNO_SP)] UNSPECV_READ_STACK_POINTER))]
  "")

;; mov m1, r0; and g2, r0 ; cmp r0, 0 ; jeq => bit g2, m1 ; jeq [r0 dead]
; see testsuite sf3296698.c
(define_peephole2
  [(set (match_operand:INTRegModes 0 "register_operand" "")
	(match_operand:<MODE> 1 "nonimmediate_operand" ""))
   (set (match_dup 0)
	(and:<MODE>
	 (match_dup 0)
	 (match_operand:<MODE> 2 "general_operand" "")))
   (set (cc0)
	(compare (match_dup 0)
		 (const_int 0)))
   (set (pc)
	(if_then_else (match_operator:<MODE> 3 "equality_operator"
		       [(cc0) (const_int 0)])
		      (match_operand 4 "" "")
		      (match_operand 5 "" "")))]
  "peep2_reg_dead_p (4, operands[0])"
  [(set (cc0)
	(compare (and:<MODE> (match_dup 1) (match_dup 2))
		 (const_int 0)))
   (set (pc)
	(if_then_else (match_operator:<MODE> 3 "equality_operator"
		       [(cc0) (const_int 0)])
		      (match_operand 4 "" "")
		      (match_operand 5 "" "")))]
  "")

